#--------------------------------------------------------------------------
#     This file is part of OASA - a free chemical python library
#     Copyright (C) 2003-2008 Beda Kosata <beda@zirael.org>

#     This program is free software; you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation; either version 2 of the License, or
#     (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.

#     Complete text of GNU GPL can be found in the file gpl.txt in the
#     main directory of the program

#--------------------------------------------------------------------------

from __future__ import division, print_function

import warnings

from math import pi, sqrt, sin, cos, atan2

#from . import graph
from . import geometry
from . import misc



def sum_of_ring_internal_angles( size):
  return (size-2)*180


def optimal_ring_iternal_angle( size, angle_shift=0):
  return sum_of_ring_internal_angles( size)/size


def gen_variations(items, n):
    if n==0:
      yield []
    else:
      for i in range(len(items)-n+1):
        for v in gen_variations(items[i+1:],n-1):
          yield [items[i]]+v


def deg_to_rad( deg):
  return pi*deg/180


def rad_to_deg( rad):
  return 180*rad/pi



class coords_optimizer(object):

  def __init__( self):
    self.cutoff_rms_grad = 1e-5
    self.cutoff_max_grad = 3e-5
    self.max_iter_number = 1000


  def optimize_coords( self, mol, bond_length=1, callback=None):
    """callback may be used to obtain information about the running optimization,
    it is called after each step with three parameters - step number, RMS grad and maxgrad"""
    self.mol = mol

    if bond_length < 0:
      bls = []
      for b in mol.edges:
        a1, a2 = b.vertices
        bls.append( sqrt( (a1.x-a2.x)**2 + (a1.y-a2.y)**2))
      self.bond_length = sum( bls) / len( bls)
    else:
      self.bond_length = bond_length

    sumdd, max_grad = self.step()
    rms_grad = sumdd / len( self.mol.vertices)
    i = 0
    ok = True
    self.end_reason = "converged"

    while rms_grad > (self.cutoff_rms_grad / self.bond_length) or max_grad > (self.cutoff_max_grad / self.bond_length):
      i += 1
      if callback:
        callback( i, rms_grad, max_grad)

      sumdd, max_grad = self.step()
      rms_grad = sumdd / len( self.mol.vertices)

      if rms_grad > 1000:
        self.end_reason = "gradien too large"
        ok = False
        break
      if i > self.max_iter_number:
        self.end_reason = "max_iter_number iterations without converged result"
        ok = False
        break

    self.i = i
    self.rms_grad = rms_grad
    self.max_grad = max_grad
    return ok


  def step( self, callback=None):

    for v in self.mol.vertices:
      v.properties_['dx'] = 0
      v.properties_['dy'] = 0

    for v in self.mol.vertices:
      self.set_gradient( v)

    dd = 0
    mdd = 0
    for v in self.mol.vertices:
      dx = v.properties_['dx']
      dy = v.properties_['dy']
      v.x += dx
      v.y += dy
      d = dx**2 + dy**2
      dd += d
      if d > mdd:
        mdd = d

    return sqrt( dd), sqrt( mdd)


  def set_gradient( self, vertex):
    dx, dy = 0, 0
    for v in vertex.neighbors:
      gx, gy = self.get_length_gradient( vertex, v)
      dx -= gx/2.0
      dy -= gy/2.0
    vertex.properties_['dx'] += dx
    vertex.properties_['dy'] += dy

    dx, dy = 0, 0
    vs = list( vertex.neighbors)
    variations = list( gen_variations( vs, 2))
    if variations:
      for var in variations:
        v1, v2 = var
        gx1, gy1, gx2, gy2 = self.get_angle_gradient( deg_to_rad( 120), vertex, v1, v2)
        v1.properties_['dx'] -= gx1/4.0
        v1.properties_['dy'] -= gy1/4.0
        v2.properties_['dx'] -= gx2/4.0
        v2.properties_['dy'] -= gy2/4.0


  def get_length_gradient( self, v1, v2, opt_length=0):
    if not opt_length:
      opt_length = self.bond_length
    dx = v1.x - v2.x
    dy = v1.y - v2.y
    dist = sqrt( dx**2 + dy**2)
    dd = dist - opt_length
    gx = (dx / dist) * dd
    gy = (dy / dist) * dd
    return gx, gy


  def get_angle_gradient( self, opt_angle, refv, v1, v2):
    expl = sqrt( 3) * self.bond_length
    gx1, gy1 = self.get_length_gradient( v1, v2, opt_length=expl)
    gx2 = -gx1
    gy2 = -gy1
    return gx1, gy1, gx2, gy2


  def get_angle_gradient2( self, opt_angle, refv, v1, v2):
    ang1 = geometry.clockwise_angle_from_east( v1.x-refv.x, v1.y-refv.y)
    ang2 = geometry.clockwise_angle_from_east( v2.x-refv.x, v2.y-refv.y)
    ang = ang1 - ang2
    sign = -geometry.on_which_side_is_point( (refv.x, refv.y, v1.x, v1.y), (v2.x, v2.y))

    while ang < 0:
      ang += 2*pi
    if ang > pi:
      ang = 2*pi - ang
      sign *= -1

    dang = (ang - opt_angle) / 2

    gx1 = (cos( ang1 - sign*dang) - cos( ang1)) * sqrt( (refv.x-v1.x)**2 + (refv.y-v1.y)**2) #self.bond_length #(v1.x-refv.x)
    gy1 = (sin( ang1 - sign*dang) - sin( ang1)) * sqrt( (refv.x-v1.x)**2 + (refv.y-v1.y)**2) #self.bond_length #(v1.y-refv.y)

    gx2 = (cos( ang2 + sign*dang) - cos( ang2)) * sqrt( (refv.x-v2.x)**2 + (refv.y-v2.y)**2) #self.bond_length #(v2.x-refv.x)
    gy2 = (sin( ang2 + sign*dang) - sin( ang2)) * sqrt( (refv.x-v2.x)**2 + (refv.y-v2.y)**2) #self.bond_length #(v2.y-refv.y)


    # control
    ang1 = geometry.clockwise_angle_from_east( v1.x-refv.x+gx1, v1.y-refv.y+gy1)
    ang2 = geometry.clockwise_angle_from_east( v2.x-refv.x+gx2, v2.y-refv.y+gy2)
    ang = ang1 - ang2
    sign = geometry.on_which_side_is_point( (refv.x, refv.y, v1.x, v1.y), (v2.x, v2.y))

    while ang < 0:
      ang += 2*pi
      #sign *= -1
    if ang > pi:
      ang = 2*pi - ang
      #sign *= -1

    dang2 = (ang - opt_angle) / 2

    if abs( dang2) > abs( dang):
      print("fuck", rad_to_deg( ang1), rad_to_deg( ang2), rad_to_deg( dang))
    else:
      #rint "good", rad_to_deg( ang1), rad_to_deg( ang2), rad_to_deg( dang)
      pass

    return gx1, gy1, gx2, gy2




if __name__ == "__main__":

  from . import smiles
  import time

  mol = smiles.text_to_mol( 'CC(C)C(C)CCCC')

  print([(v.x,v.y) for v in mol.vertices])

  opt = coords_optimizer()
  t = time.time()
  ok = opt.optimize_coords( mol, bond_length=0.9)
  print("converged:", ok)
  print("%d iterations, RMS grad %f, max grad %f" % (opt.i, opt.rms_grad, opt.max_grad))
  print("time %.1f ms" % ((time.time() - t) * 1000))


  print([(v.x,v.y) for v in mol.vertices])

  from . import coords_generator

  coords_generator.show_mol( mol)



#--------------------------------------------------------------------------
#     This file is part of OASA - a free chemical python library
#     Copyright (C) 2003-2008 Beda Kosata <beda@zirael.org>

#     This program is free software; you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation; either version 2 of the License, or
#     (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.

#     Complete text of GNU GPL can be found in the file gpl.txt in the
#     main directory of the program

#--------------------------------------------------------------------------

import copy
import math
import xml.dom.minidom as dom

from . import misc
from . import geometry
from . import transform
from . import dom_extensions



class svg_out(object):

  show_hydrogens_on_hetero = False
  margin = 15
  line_width = 2
  bond_width = 6
  # should individual parts of an edge be grouped together
  group_items = True


  def __init__( self):
    pass

  def mol_to_svg( self, mol, before=None, after=None):
    """before and after should be methods or functions that will take one
    argument - svg_out instance and do whatever it wants with it - usually
    adding something to the resulting DOM tree"""
    self.document = dom.Document()
    top = dom_extensions.elementUnder( self.document,
                                       "svg",
                                       attributes=(("xmlns", "http://www.w3.org/2000/svg"),
                                                   ("version", "1.0")))
    self.top = dom_extensions.elementUnder( top, "g",
                                            attributes=(("stroke", "#000"),
                                                        ("stroke-width", "1.0")))

    x1, y1, x2, y2 = None, None, None, None
    for v in mol.vertices:
      if x1 is None or x1 > v.x:
        x1 = v.x
      if x2 is None or x2 < v.x:
        x2 = v.x
      if y1 is None or y1 > v.y:
        y1 = v.y
      if y2 is None or y2 < v.y:
        y2 = v.y

    w = int( x2 - x1 + 2*self.margin)
    h = int( y2 - y1 + 2*self.margin)
    top.setAttribute( "width", str( w))
    top.setAttribute( "height", str( h))

    self.transformer = transform.transform()
    self.transformer.set_move( -x1+self.margin, -y1+self.margin)

    self.molecule = mol

    if before:
      before( self)

    for e in copy.copy( mol.edges):
      self._draw_edge( e)
    for v in mol.vertices:
      self._draw_vertex( v)

    if after:
      after( self)

    return self.document


  def paper_to_canvas_coord( self, x):
    return x


  def prepare_dumb_transformer( self):
    tr = transform.transform()
    tr.set_scaling( self.paper_to_canvas_coord( 1))
    return tr


  def _create_parent( self, item, top):
    if self.group_items:
      parent = dom_extensions.elementUnder( top, "g")
      if 'svg_id' in item.properties_:
        parent.setAttribute( "id", item.properties_['svg_id'])
    else:
      parent = self.top
    return parent


  def _draw_edge( self, e):
    v1, v2 = e.vertices
    start = self.transformer.transform_xy( v1.x, v1.y)
    end = self.transformer.transform_xy( v2.x, v2.y)
    parent = self._create_parent( e, self.top)

    if e.order == 1:
      self._draw_line( parent, start, end, line_width=self.line_width)

    if e.order == 2:
      side = 0
      # find how to center the bonds
      # rings have higher priority in setting the positioning
      for ring in self.molecule.get_smallest_independent_cycles():
        if v1 in ring and v2 in ring:
          side += sum(geometry.on_which_side_is_point(start + end, self.transformer.transform_xy(a.x, a.y))
                          for a in ring
                              if a != v1 and a != v2)
      # if rings did not decide, use the other neigbors
      if not side:
        for v in v1.neighbors + v2.neighbors:
          if v != v1 and v!= v2:
            side += geometry.on_which_side_is_point( start+end, (self.transformer.transform_xy( v.x, v.y)))
      if side:
        self._draw_line( parent, start, end, line_width=self.line_width)
        x1, y1, x2, y2 = geometry.find_parallel( start[0], start[1], end[0], end[1], self.bond_width*misc.signum( side))
        self._draw_line( parent, (x1, y1), (x2, y2), line_width=self.line_width)
      else:
        for i in (1,-1):
          x1, y1, x2, y2 = geometry.find_parallel( start[0], start[1], end[0], end[1], i*self.bond_width*0.5)
          self._draw_line( parent, (x1, y1), (x2, y2), line_width=self.line_width)


    elif e.order == 3:
      self._draw_line( parent, start, end, line_width=self.line_width)
      for i in (1,-1):
        x1, y1, x2, y2 = geometry.find_parallel( start[0], start[1], end[0], end[1], i*self.bond_width*0.7)
        self._draw_line( parent, (x1, y1), (x2, y2), line_width=self.line_width)


  def _draw_vertex( self, v):
    parent = self._create_parent( v, self.top)

    if v.symbol != "C" or v.charge != 0 or v.multiplicity != 1:
      x = v.x - 5
      y = v.y + 6
      x1 = x
      x2 = x + 12
      y1 = y - 12
      y2 = y + 2
      text = v.symbol
      if self.show_hydrogens_on_hetero:
        if v.free_valency == 1:
          text += "H"
        elif v.free_valency > 1:
          text += "H%d" % v.free_valency
      # charge
      if v.charge == 1:
        text += "+"
      elif v.charge == -1:
        text += "-"
      elif v.charge > 1:
        text += str( v.charge) + "+"
      elif v.charge < -1:
        text += str( v.charge)
      # radicals
      if v.multiplicity in (2,3):
        self._draw_circle( parent, self.transformer.transform_xy((x2+x1)/2,y-17), fill_color="#000", opacity=1, radius=3)
        if v.multiplicity == 3:
          self._draw_circle( parent, self.transformer.transform_xy((x2+x1)/2,y+5), fill_color="#000", opacity=1, radius=3)
      self._draw_rectangle( parent, self.transformer.transform_4( (x1, y1, x2, y2)), fill_color="#fff")
      self._draw_text( parent, self.transformer.transform_xy(x,y), text)


  def _draw_line( self, parent, start, end, line_width=1, capstyle=""):
    x1, y1 = start
    x2, y2 = end
    line = dom_extensions.elementUnder( parent, 'line',
                                        (( 'x1', str( x1)),
                                         ( 'y1', str( y1)),
                                         ( 'x2', str( x2)),
                                         ( 'y2', str( y2))))


  def _draw_text( self, parent, xy, text, font_name="Arial", font_size=16):
    x, y = xy
    dom_extensions.textOnlyElementUnder( parent, "text", text,
                                         (( "x", str( x)),
                                          ( "y", str( y)),
                                          ( "font-family", font_name),
                                          ( "font-size", str( font_size)),
                                          ( 'fill', "#000")))


  def _draw_rectangle( self, parent, coords, fill_color="#fff", stroke_color="#fff"):
    x, y, x2, y2 = coords
    dom_extensions.elementUnder( parent, 'rect',
                                 (( 'x', str( x)),
                                  ( 'y', str( y)),
                                  ( 'width', str( x2-x)),
                                  ( 'height', str( y2-y)),
                                  ( 'fill', fill_color),
                                  ( 'stroke', stroke_color)))

  def _draw_circle( self, parent, xy, radius=5, fill_color="#fff", stroke_color="#fff", id="", opacity=0):
    x, y = xy
    el = dom_extensions.elementUnder( parent, 'ellipse',
                                      (( 'cx', str( x)),
                                       ( 'cy', str( y)),
                                       ( 'rx', str( radius)),
                                       ( 'ry', str( radius)),
                                       ( 'stroke-width', "1"),
                                       ( 'fill', fill_color),
                                       ( 'stroke', stroke_color),
                                       ( 'fill-opacity', str(opacity)),
                                       ( 'stroke-opacity', str(opacity)),
                                      ))
    if id:
      el.setAttribute( "id", id)


def mol_to_svg( mol, filename):
  c = svg_out()
  tree = c.mol_to_svg( mol)
  with open(filename, 'wb') as f:
    f.write(tree.toxml('utf-8'))


if __name__ == "__main__":

  #from . import inchi
  #mol = inchi.text_to_mol( "1/C7H6O2/c8-7(9)6-4-2-1-3-5-6/h1-5H,(H,8,9)", include_hydrogens=False, calc_coords=30)
  #from . import smiles
  #mol = smiles.text_to_mol( "CC[CH]", calc_coords=40)
  from . import molfile
  mol = molfile.file_to_mol(open("/home/beda/bkchem/bkchem/untitled0.mol", "r"))
  mol_to_svg( mol, "output.svg")




#--------------------------------------------------------------------------
#     This file is part of OASA - a free chemical python library
#     Copyright (C) 2003-2008 Beda Kosata <beda@zirael.org>

#     This program is free software; you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation; either version 2 of the License, or
#     (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.

#     Complete text of GNU GPL can be found in the file gpl.txt in the
#     main directory of the program

#--------------------------------------------------------------------------

from __future__ import print_function

import re

from . import misc
from . import smiles
from . import coords_generator
from .config import Config
from .known_groups import name_to_smiles
from .periodic_table import periodic_table
from .oasa_exceptions import oasa_invalid_atom_symbol



class linear_formula( object):

  def __init__( self, text="", start_valency=0, end_valency=0, mol=None):
    """valency specifies the already occupied valency of the submited formula,
    it is usually used when parsing groups"""
    self.molecule = None
    if text:
      self.parse_text( text, start_valency=start_valency, end_valency=end_valency, mol=mol)


  def parse_text( self, text, start_valency=0, end_valency=0, mol=None):
    text = self.expand_abbrevs( text)
    mol = self.parse_form( text, start_valency=start_valency, mol=mol, reverse=False)
    if mol:
      # are there any atoms?
      if not mol.vertices:
        return None

      # now we check if the structure is complete
      for v in mol.vertices:
        if v.free_valency:
          # we permit free valency at the last atom if end_valency was given
          if not (v == self.get_last_free_atom(mol) and v.free_valency == end_valency):
            return None

      if start_valency:
        mol.remove_vertex( mol.vertices[0]) # remove the dummy

      # are there any atoms, again?
      if not mol.vertices:
        return None

      # is the molecule connected?
      if not mol.is_connected():
        return None

      # remove hydrogens, but only if the molecule is not completely made of them :)
      mol.remove_unimportant_hydrogens()

      # set some objects attributes for later reading
      self.molecule = mol
      if start_valency:
        self.first_atom = mol.vertices[0]
      else:
        self.first_atom = None
      if end_valency:
        self.last_atom = self.get_last_free_atom(mol)
      else:
        self.last_atom = None
      return mol


  def parse_form( self, text, start_valency=0, mol=None, reverse=False):
    form = text

    # the code itself
    if not mol:
      mol = Config.create_molecule()

    # create the dummy atom
    if start_valency:
      last_atom = mol.create_vertex()
      last_atom.valency = start_valency
      mol.add_vertex( last_atom)
    else:
      last_atom = None

    # check if there are branches in the formula
    if "(" not in form:
      # there are no subbranches
      chunks = re.split( "([A-Z][a-z]?[0-9]?[+-]?)", form)
      if reverse:
        chunks = reverse_chunks( chunks)
      for chunk in chunks:
        if chunk:
          atms = self.chunk_to_atoms( chunk, mol)
          if atms is None:
            return None
          last_atom = self.get_last_free_atom( mol)
          for a in atms:
            mol.add_vertex( a)
            if last_atom:
              max_val = min( last_atom.free_valency, a.free_valency, 3)
              if max_val <= 0 and last_atom.free_valency <= 0:
                if last_atom.raise_valency():
                  max_val = min( last_atom.free_valency, a.free_valency, 3)
              b = mol.create_edge()
              b.order = max_val
              mol.add_edge( last_atom, a, b)
            else:
              last_atom = a
    else:
      for chunk, count in gen_formula_fragments( form, reverse=reverse):
        if chunk:
          last_atom = self.get_last_free_atom( mol)
          do_linear = False # should we string the fragments rather than adding them all to the last atom
          if last_atom and last_atom.free_valency < count:
            do_linear = True

          for j in range( count):
            if chunk[0] == "!":
              # the form should be a smiles
              m = smiles.text_to_mol( chunk[1:], calc_coords=0)
              m.add_missing_hydrogens()
              hs = [v for v in m.vertices[0].neighbors if v.symbol == 'H']
              m.disconnect( hs[0], m.vertices[0])
              m.remove_vertex( hs[0])
              smile = True
            else:
              val = last_atom and 1 or 0
              m = self.parse_form( chunk, start_valency=val, mol=mol.create_graph())
              smile = False
            if not m:
              return None

            if not last_atom:
              # !!! this should not happen in here
              mol.insert_a_graph( m)
              # if there are multiple chunks without a previous atom, we should create a linear fragment
              #last_atom = [v for v in mol.vertices if v.free_valency > 0][-1]
            else:
              if not smile:
                m.remove_vertex( m.vertices[0]) # remove the dummy
              mol.insert_a_graph( m)
              b = mol.create_edge()
              mol.add_edge( last_atom, m.vertices[0], b)
              if do_linear:
                last_atom = m.vertices[0]

    return mol



  def chunk_to_atoms( self, chunk, mol):
    m = re.match( "([A-Z][a-z]?)([0-9])?([+-])?", chunk)
    if m:
      name = m.group( 1)
      number = m.group( 2) and int( m.group(2)) or 1
      sign = m.group( 3) and (int( m.group(3)+'1')) or 0
      ret = []
      for i in range( number):
        v = mol.create_vertex()
        try:
          v.symbol = name
        except oasa_invalid_atom_symbol:
          return None
        v.charge = sign
        ret.append( v)
      return ret


  def get_last_free_atom( self, mol):
    # check if there is something with a free valency
    if not mol.vertices:
      return None
    atoms = [o for o in misc.reverse( mol.vertices)]
    for a in atoms:
      if a.free_valency > 0:
        return a
    # if its not the case
    for i, a in enumerate( atoms[0:-1]):
      b = a.get_edge_leading_to( atoms[i+1])
      if b and b.order > 1:
        b.order -= 1
        return a
    # well, we cannot do anything else
    return atoms[-1]


  def expand_abbrevs( self, text):
    # at first sort the text according to length, so that the longest are expanded first
    # (MMTr and not Tr)
    keys = sorted([(len( k), k) for k in name_to_smiles.keys()], reverse=True)
    for l, key in keys:
      val = name_to_smiles[ key]
      text = text.replace( key, "(!%s)" % val)
    return text


def reverse_chunks( chunks):
  chunks.reverse()
  for i in range( 0, len( chunks), 2):
    chunks[i], chunks[i+1] = chunks[i+1],chunks[i]
  return chunks


def gen_formula_fragments( formula, reverse=False):
  chunks = list( gen_formula_fragments_helper( formula))
  if reverse:
    chunks.reverse()
  i = 0
  while i < len( chunks):
    chunk, brack = chunks[ i]
    if brack and i < len( chunks) - 1:
      next, nbrack = chunks[i+1]
      if not nbrack:
        count, rest = split_number_and_text( next)
        if count is None:
          yield chunk, 1
        else:
          chunks[i+1] = (rest, nbrack)
          if not rest:
            i += 1
          yield chunk, count
      else:
        yield chunk, 1
    else:
      yield chunk, 1
    i += 1


def gen_formula_fragments_helper( formula):
  opened_brackets = 0
  to_ret = []
  for ch in formula:
    if ch not in "()":
      to_ret.append( ch)
    elif ch == "(":
      if opened_brackets == 0:
        if to_ret:
          yield ''.join( to_ret), False
          to_ret = []
      else:
        to_ret.append( "(")
      opened_brackets += 1
    elif ch == ")":
      opened_brackets -= 1
      if opened_brackets < 0:
        opened_brackets = 0
      if opened_brackets == 0:
        if to_ret:
          yield ''.join( to_ret), True
          to_ret = []
      else:
        to_ret.append( ")")
  if to_ret:
    yield ''.join( to_ret), False



def split_number_and_text( txt):
  last = None
  for i in range( len( txt)):
    try:
      last = int( txt[0:i+1])
    except ValueError:
      return last, txt[i:]
  return last, ""



def reverse_formula( text):
  all_chunks = []
  form = text
  if "(" not in form:
    chunks = re.split( "([A-Z][a-z]?[0-9]?[+-]?)", form)
    for chunk in chunks:
      if chunk:
        all_chunks.append( chunk)
  else:
    for chunk, count in gen_formula_fragments_helper( form):
      if count:
        all_chunks.append(")")
      if chunk:
        all_chunks.extend( reverse_formula( chunk))
      if count:
        all_chunks.append("(")
#
#        if count:
#          all_chunks.append( str(count))

  print(all_chunks)
  #all_chunks.reverse()
  return all_chunks


if __name__ == "__main__":
  form = 'CH3(CH2)7'
  #form = 'CH3((CH2)2)2O'
  form = "1) Bu"

  a = linear_formula( form, start_valency=0, end_valency=1)
  if a.molecule:
    m = a.molecule
    coords_generator.calculate_coords( m)
    coords_generator.show_mol( m)
  else:
    print("nothing")




#--------------------------------------------------------------------------
#     This file is part of OASA - a free chemical python library
#     Copyright (C) 2003-2008 Beda Kosata <beda@zirael.org>

#     This program is free software; you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation; either version 2 of the License, or
#     (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.

#     Complete text of GNU GPL can be found in the file gpl.txt in the
#     main directory of the program

#--------------------------------------------------------------------------



class diedge(object):

  def __init__(self, vs=[]):
    self.vertices = []
    self.set_vertices(vs)
    self.properties_ = {}


  def __str__(self):
    return "Directed edge between %s %s" % tuple(map(str, self.vertices))


  def set_vertices(self, vs=[]):
    if vs and len(vs) == 2:
      self.vertices = vs


  def get_vertices(self):
    return self.vertices


  @property
  def neighbor_edges(self):
    v1, v2 = self.vertices
    out1 = [e for e in v1.neighbor_edges if e != self]
    out2 = [e for e in v2.neighbor_edges if e != self]
    return out1 + out2


  def get_neighbor_edges2(self):
    """Return 2 lists of neighbor edges (one for one side, one for the other).

    """
    v1, v2 = self.vertices
    out1 = [e for e in v1.neighbor_edges if e != self]
    out2 = [e for e in v2.neighbor_edges if e != self]
    return out1, out2




#--------------------------------------------------------------------------
#     This file is part of OASA - a free chemical python library
#     Copyright (C) 2003-2008 Beda Kosata <beda@zirael.org>

#     This program is free software; you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation; either version 2 of the License, or
#     (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.

#     Complete text of GNU GPL can be found in the file gpl.txt in the
#     main directory of the program

#--------------------------------------------------------------------------

"""Abstract parent class for all converters.

"""



class converter_base(object):

  STATUS_OK = 1
  STATUS_CRITICAL_ERROR = 2

  # standard converter attrs
  reads_text = False
  writes_text = False
  reads_files = False
  writes_files = False

  default_configuration = {}

  def __init__( self):
    self.configuration = {}
    for k, v in self.default_configuration.items():
      self.configuration[k] = v
    self.cleanup()


  def clean_logs( self):
    # here all warnings for the last run should be stored
    self.warnings = []
    # here error messages for the last run will be stored
    self.errors = []


  def mols_to_text( self, structures):
    self.clean_logs()
    pass


  def read_text( self, text):
    self.clean_logs()
    pass


  def mols_to_file( self, structures, f):
    self.clean_logs()
    pass


  def read_file( self, f):
    self.clean_logs()
    pass


  def cleanup( self):
    self.clean_logs()
    self.result = []
    self.last_status = None




#--------------------------------------------------------------------------
#     This file is part of OASA - a free chemical python library
#     Copyright (C) 2003-2008 Beda Kosata <beda@zirael.org>

#     This program is free software; you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation; either version 2 of the License, or
#     (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.

#     Complete text of GNU GPL can be found in the file gpl.txt in the
#     main directory of the program

#--------------------------------------------------------------------------

from . import oasa_exceptions



class stereochemistry(object):

  def __init__( self, center=None, references=None, value=None):
    self.center = center
    if not references:
      self.references = []
    else:
      self.references = references
    self.value = value


  @property
  def value(self):
    return self._value


  @value.setter
  def value(self, value):
    self._value = value


  @property
  def references(self):
    return self._references


  @references.setter
  def references(self, references):
    self._references = references


  @property
  def center(self):
    return self._center


  @center.setter
  def center(self, center):
    self._center = center



class cis_trans_stereochemistry(stereochemistry):

  UNDEFINED = 0
  OPPOSITE_SIDE = 1
  SAME_SIDE = 2

  # Override value
  @property
  def value(self):
    return stereochemistry.value.__get__(self)


  @value.setter
  def value(self, value):
    if value not in (self.SAME_SIDE, self.OPPOSITE_SIDE, self.UNDEFINED):
      raise oasa_exceptions.oasa_stereochemistry_error( "invalid stereochemistry identifier '%s'" % value)
    stereochemistry.value.__set__(self, value)


  # Override references
  @property
  def references(self):
    return stereochemistry.references.__get__(self)


  @references.setter
  def references(self, references):
    if len( references) != 4:
      raise oasa_exceptions.oasa_stereochemistry_error( "wrong number of references in stereochemistry specification '%s'" % len( references))
    stereochemistry.references.__set__(self, references)


  def get_other_end( self, ref):
    if not ref in self.references:
      raise ValueError("Submitted object is not referenced in this stereochemistry object.")
    ref1, _r1, _r2, ref2 = self.references
    return ref is ref1 and ref2 or ref1



class tetrahedral_stereochemistry(stereochemistry):

  UNDEFINED = 0
  # as in CIP - put the last atom behind the reference and observer the first 3
  # this should create the same result as SMILES notation - put firts in front and observe the rest
  CLOCKWISE = 1
  ANTICLOCKWISE = 2

  # Override value
  @property
  def value(self):
    return stereochemistry.value.__get__(self)


  @value.setter
  def value(self, value):
    if value not in (self.CLOCKWISE, self.ANTICLOCKWISE, self.UNDEFINED):
      raise oasa_exceptions.oasa_stereochemistry_error( "invalid stereochemistry identifier '%s'" % value)
    stereochemistry.value.__set__(self, value)


  # Override references
  @property
  def references(self):
    return stereochemistry.references.__get__(self)


  @references.setter
  def references(self, references):
    if len( references) != 4:
      raise oasa_exceptions.oasa_stereochemistry_error( "wrong number of references in stereochemistry specification '%s'" % len( references))
    stereochemistry.references.__set__(self, references)



class explicit_hydrogen(object):
  """Placeholder for explicit hydrogen in stereochemistry references.

  """
  def __eq__(self, other):
    if isinstance(other, explicit_hydrogen):
      return True
    else:
      return False




#--------------------------------------------------------------------------
#     This file is part of OASA - a free chemical python library
#     Copyright (C) 2003-2008 Beda Kosata <beda@zirael.org>

#     This program is free software; you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation; either version 2 of the License, or
#     (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.

#     Complete text of GNU GPL can be found in the file gpl.txt in the
#     main directory of the program

#--------------------------------------------------------------------------

import sys
import copy

from warnings import warn

from . import graph
from . import periodic_table as PT
from .common import is_uniquely_sorted



class chem_vertex(graph.vertex):
  """Parent class of atoms, groups etc.

  It defines common properties for vertices used in chemical context.
  It should not be instantiated directly, but rather inherited from.
  """
  attrs_to_copy = graph.vertex.attrs_to_copy + ("charge","x","y","z","multiplicity","valency","charge","free_sites")

  def __init__( self, coords=None):
    graph.vertex.__init__( self)
    self.charge = 0
    self.free_sites = 0
    # None means not set (used)
    if coords:
      self.x, self.y, self.z = coords
    else:
      self.x = None
      self.y = None
      self.z = None
    self._multiplicity = 1


  def matches( self, other):
    if other is self:
      return False
    return True


  @property
  def coords(self):
    """Atom coordinates.

    """
    return self.x, self.y, self.z


  @coords.setter
  def coords(self, coords):
    if len( coords) == 2:
      self.x, self.y = coords
      self.z = 0
    elif len( coords) == 3:
      self.x, self.y, self.z = coords
    else:
      raise Exception("wrong number of coordinates")


  @property
  def charge(self):
    """Atom charge.

    """
    return self._charge


  @charge.setter
  def charge(self, charge):
    self._clean_cache()
    self._charge = charge


  @property
  def multiplicity(self):
    """Atom multiplicity.

    """
    return self._multiplicity


  @multiplicity.setter
  def multiplicity(self, multiplicity):
    self._clean_cache()
    self._multiplicity = multiplicity


  @property
  def valency(self):
    """Atom valency.

    """
    return self._valency


  @valency.setter
  def valency(self, valency):
    self._clean_cache()
    self._valency = valency


  @property
  def occupied_valency(self):
    """Atom's occupied valency.

    """
    i = 0
    for b in self._neighbors.keys():
      ord = b.order
      if ord == 4:
        ord = 1
      i += ord
    return i


  @property
  def free_valency(self):
    """Atom's free valency.

    """
    try:
      return self._cache[ 'free_valency']
    except KeyError:
      x = self.valency - self.occupied_valency
      self._cache[ 'free_valency'] = x
      return x


  @property
  def weight(self):
    """Atom weight.

    """
    try:
      return PT.periodic_table[self.symbol]['weight']
    except:
      return 0


  @property
  def free_sites(self):
    """Atom's free sites.

    """
    really_free = self._free_sites - self.occupied_valency
    if really_free < 0:
      return 0
    else:
      return really_free


  @free_sites.setter
  def free_sites(self, free_sites):
    self._free_sites = free_sites


  def get_x( self):
    return self.x or 0


  def get_y( self):
    return self.y or 0


  def get_z( self):
    return self.z or 0


  def has_aromatic_bonds( self):
    for b in self._neighbors.keys():
      if b.aromatic:
        return 1
    return 0


  def bond_order_changed( self):
    """called by a bond when its order was changed"""
    self._clean_cache()


  def get_hydrogen_count( self):
    return 0




#--------------------------------------------------------------------------
#     This file is part of OASA - a free chemical python library
#     Copyright (C) 2003-2008 Beda Kosata <beda@zirael.org>

#     This program is free software; you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation; either version 2 of the License, or
#     (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.

#     Complete text of GNU GPL can be found in the file gpl.txt in the
#     main directory of the program

#--------------------------------------------------------------------------

import sys
if not (sys.version_info[0] > 2 or (sys.version_info[0] == 2 and sys.version_info[1] >= 6)):
  raise ImportError("Python version %d.%d is lower than 2.6 which is needed by OASA" % sys.version_info[0:2])


from . import atom
from . import bond
from . import molecule
from . import smiles
from . import coords_generator
from . import coords_optimizer
from . import molfile
from . import inchi
from . import cdml
from . import graph
from . import linear_formula
from . import periodic_table
from . import config
from . import query_atom
from . import chem_vertex
from . import oasa_exceptions
from . import subsearch
from . import svg_out
from . import stereochemistry
from . import geometry
from . import transform3d
from . import transform
from . import known_groups

atom = atom.atom
bond = bond.bond
molecule = molecule.molecule
query_atom = query_atom.query_atom
chem_vertex = chem_vertex.chem_vertex

allNames = ['atom', 'bond', 'chem_vertex', 'coords_generator', 'config',
            'coords_optimizer', 'geometry', 'graph', 'inchi', 'known_groups',
            'linear_formula', 'molecule', 'molfile', 'name_database',
            'oasa_exceptions', 'periodic_table', 'query_atom', 'smiles',
            'stereochemistry', 'subsearch', 'svg_out', 'transform',
            'transform3d']

try:
  from . import cairo_out
except ImportError:
  CAIRO_AVAILABLE = False
else:
  allNames.append("cairo_out")
  CAIRO_AVAILABLE = True

# inchi_key
try:
  from . import inchi_key
except Exception as e:
  #print >> sys.stderr, "Module inchi_key could not be loaded - inchi_key related features will be disabled\nSee the error message for more info:\n  %s" % e
  INCHI_KEY_AVAILABLE = False
else:
  allNames.append("inchi_key")
  INCHI_KEY_AVAILABLE = True

# name_database (requires inchi_key which requires mhash in Python 2.4)
try:
  from . import name_database
except Exception as e:
  NAME_DATABASE_AVAILABLE = False
else:
  allNames.append("name_database")
  NAME_DATABASE_AVAILABLE = True

# structure_database requires sqlite
try:
  from . import structure_database
except Exception as e:
  #print >> sys.stderr, "Module structure_database could not be loaded - structure_database related features will be disabled\nSee the error message for more info:\n  %s" % e
  STRUCTURE_DATABASE_AVAILABLE = False
else:
  allNames.append("structure_database")
  STRUCTURE_DATABASE_AVAILABLE = True

# pybel
try:
  from . import pybel_bridge
except Exception as e:
  #print >> sys.stderr, "The 'pybel_bridge' python module could not be loaded - oasa-pybel integration will be disabled\nSee the error message for more info:\n  %s" % e
  PYBEL_AVAILABLE = False
else:
  allNames.append("pybel_bridge")
  PYBEL_AVAILABLE = True


__all__ = allNames




#--------------------------------------------------------------------------
#     This file is part of OASA - a free chemical python library
#     Copyright (C) 2003-2008 Beda Kosata <beda@zirael.org>

#     This program is free software; you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation; either version 2 of the License, or
#     (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.

#     Complete text of GNU GPL can be found in the file gpl.txt in the
#     main directory of the program

#--------------------------------------------------------------------------

"""The bond resides here.

"""

##  the order 1,2,3 is self-evident, order=4 means aromatic bond for which no
##  localized bond order is available


import sys
sys.path.append( '../')
import math

from warnings import warn

from . import graph



class bond(graph.edge, object):
  """Based on edge, however the vertices are not a Set anymore.

  We need to preserve the order (for instance for wedge bonds).
  Type is:
  'n' - normal
  'w' - wedge
  'h' - hatch
  'a' - adder
  'b' - bold
  'd' - dash
  """
  attrs_to_copy = graph.edge.attrs_to_copy + ("order","aromatic","type")

  def __init__( self, vs=[], order=1, type='n'):
    graph.edge.__init__( self, vs=vs)
    self.set_vertices( vs)
    self.aromatic = None  # None means it was not set
    self.order = order
    self.type = type
    self.properties_ = {}
    self.stereochemistry = None


  def __str__( self):
    if self._vertices:
      return "bond between %s %s" % tuple( map( str, self.vertices))
    else:
      return "bond, no vertices set"


  def matches( self, other):
    if self.order == other.order:
      return True
    return False


  @property
  def vertices(self):
    """Tuple of 2 vertices (start, end).

    """
    return self._vertices


  @vertices.setter
  def vertices(self, vs=[]):
    """Sets the vertices this edge connects.

    """
    assert len( vs) == 2 or len( vs) == 0
    #if len( vs) == 2 and vs[0] == vs[1]:
    #  warn( "creating bond with both ends equal", UserWarning, 2)
    self._vertices = list( vs)


  @property
  def order(self):
    """Bond order.

    1-3 for normal bonds,
    4   for aromatic bonds for which localized order is not available,
    for localized aromatic bonds check the bond.aromatic boolean attribute.
    """
    if self._order is None and self.aromatic:
      return 4
    else:
      return self._order


  @order.setter
  def order(self, order):
    [a.bond_order_changed() for a in self.vertices]
    if order == 4:
      self._order = None
      self.aromatic = 1
    else:
      self._order = order
      #self.aromatic = None


  @property
  def length(self):
    """Bond length.

    """
    if len( self.vertices) == 2:
      v1, v2 = self.vertices
      return math.sqrt( (v1.x-v2.x)**2 + (v1.y-v2.y)**2)
    else:
      return 0


### TODO

# support for the bond.aromatic



#--------------------------------------------------------------------------
#     This file is part of OASA - a free chemical python library
#     Copyright (C) 2003-2008 Beda Kosata <beda@zirael.org>

#     This program is free software; you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation; either version 2 of the License, or
#     (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.

#     Complete text of GNU GPL can be found in the file gpl.txt in the
#     main directory of the program

#--------------------------------------------------------------------------

from __future__ import print_function

import os

from . import smiles
from .graph.digraph import digraph



class substructure_search_manager( object):

  substructure_def_file = os.path.join( os.path.dirname( __file__), "subsearch_data.txt")
  ring_def_file = os.path.join( os.path.dirname( __file__), "subsearch_rings.txt")

  def __init__( self):
    self.structures = digraph()  # graph describing the relations between individual structures
    self.search_trees = []  # list of substructure instances that for a tree
    self.rings = {}
    self.fill_data()

  def fill_data( self):
    from . import subsearch_data
    for struct in subsearch_data.structures:
      compound_type, name, smiles_string, to_ignore = struct
      sub = substructure( name, compound_type, smiles=smiles_string, atoms_to_ignore=to_ignore)
      v = self.structures.create_vertex()
      v.value = sub
      self.structures.add_vertex( v)
    for ring_desc in subsearch_data.rings:
      name, smiles_string, ring_hash = ring_desc
      rng = ring( name, smiles_string, ring_hash=ring_hash)
      self.rings[ ring_hash] = rng
    # the following takes some time
    self._analyze_structure_dependencies()

  def _read_structure_file( self, name=""):
    """may be used to read data directly from source txt files.
    is deprecated in favor of automatically build subsearch_data.py module"""
    with open(name or self.substructure_def_file, 'r') as f:
      for line in f:
        l = line.strip()
        if l and not l.startswith("#"):
          parts = l.split(";")
          if len(parts) < 3:
            raise ValueError("Wrong line in data file: '%s'" % l)
          compound_type, name, smiles_string = [x.strip() for x in parts[:3]]
          if len(parts) > 3:
            to_ignore = parts[3].strip()
          else:
            to_ignore = ""
          to_ignore = map(int, filter(None, to_ignore.split(",")))
          if not name.strip():
            name = compound_type
          sub = substructure(name,
                             compound_type,
                             smiles=smiles_string,
                             atoms_to_ignore=to_ignore)
          v = self.structures.create_vertex()
          v.value = sub
          self.structures.add_vertex(v)
    self._analyze_structure_dependencies()

  def _read_ring_file( self, name=""):
    """may be used to read data directly from source txt files.
    is deprecated in favor of automatically build subsearch_data.py module"""
    with open(name or self.ring_def_file, 'r') as f:
      for line in f:
        l = line.strip()
        if l and not l.startswith("#"):
          parts = [x.strip() for x in l.split(";")]
          if len(parts) != 3:
            raise ValueError("Wrong line in data file: '%s'" % l)
          name, smiles_string, ring_hash = parts
          rng = ring(name, smiles_string, ring_hash=ring_hash)
          self.rings[ring_hash] = rng

  def _analyze_structure_dependencies( self):
    for v1 in self.structures.vertices:
      for v2 in self.structures.vertices:
        if v1 is not v2:
          sub1 = v1.value
          sub2 = v2.value
          if sub1.structure.contains_substructure( sub2.structure):
            self.structures.add_edge( v2, v1)

  def _find_head_structures( self):
    for v1 in self.structures.vertices:
      v1.properties_['in_links'] = []
      for v2 in self.structures.vertices:
        if v1 in v2.neighbors:
          v1.properties_['in_links'].append( v2.value)
    heads = []
    for v in self.structures.vertices:
      if len( v.properties_['in_links']) == 0:
        # we have got a top of the tree
        heads.append( v)
    return heads

  def _compute_search_trees( self):
    heads = self._find_head_structures()
    assert heads
    for head in heads:
      d = self.structures.mark_vertices_with_distance_from( head)
      dv = [(v.properties_['d'],v) for v in self.structures.vertices if 'd' in v.properties_]
      dv.sort( reverse=True)
      for d,v in dv:
        for d,parent in dv:
          if parent.value in v.properties_['in_links']:
            if v.value not in parent.value.children:
              # this ensures only one copy of a substructure in a tree
              parent.value.children.append( v.value)
            break
    return [h.value for h in heads]

  def which_substructure_is_more_specific( self, s1, s2):
    """returns 1 if s1 is more specific,
               2 if s2 is more specific,
               3 if both are the same,
               4 if the relationship cannot be established
               (there is not connection between s1 and s2 in the
               substructure interdependency graph - this signifies
               a problem, probably in the matching algorithm
               """
    if s1 is s2:
      return 3
    # just find vertices corresponding to s1 and s2
    v1 = None
    v2 = None
    for v in self.structures.vertices:
      if v.value == s1:
        v1 = v
      elif v.value == s2:
        v2 = v
    assert v1
    assert v2
    p1 = self.structures.path_exists( v1, v2)
    p2 = self.structures.path_exists( v2, v1)
    if (p1 and p2) or (not p1 and not p2):
      # both paths exist or none does - this should not happen
      return 4
    if p1:
      return 2
    if p2:
      return 1

  def find_substructures_in_mol( self, mol):
    # get the hits
    hits2 = []
    for v in self.structures.vertices:
      struct = v.value
      ms = struct.find_matches( mol)
      hits2 += ms
    # weed out the hits that match inside a ring
    ring_hits = self.find_rings_in_mol( mol)
    hits = []
    for hit in hits2:
      keep = True
      for rhit in ring_hits:
        if set(rhit.get_significant_atoms()) & set(hit.get_significant_atoms()):
          keep = False
          break
      if keep:
        hits.append( hit)
    # weed out overlapping hits - leave only the most significant ones
    hit_num = len( hits)
    to_delete = True
    while to_delete:
      to_delete = []
      for i,hit1 in enumerate( hits):
        for hit2 in hits[i+1:]:
          if hit1 is not hit2:
            if set(hit1.get_significant_atoms()) & set(hit2.get_significant_atoms()):
              winner = self.which_substructure_is_more_specific( hit1.substructure, hit2.substructure)
              if winner == 1:
                to_delete.append( hit2)
              elif winner == 2:
                to_delete.append( hit1)
              elif winner == 3:
                pass # we preserve both hits
              else:
                raise ValueError( "Relationship between competing fragments could not be established,\nthere is probaly an error in the substructure matching code.")
        if to_delete:
          break
      hits = [hit for hit in hits if not hit in to_delete]
    return ring_hits + hits

  def find_rings_in_mol( self, mol, detect_aromatic=True):
    hits = []
    if detect_aromatic:
      # this is necessary to correctly process fused aromatic rings
      # with improperly localized bonds
      mol.mark_aromatic_bonds()
    for ering in mol.get_smallest_independent_cycles_e():
      vring = mol.edge_subgraph_to_vertex_subgraph( ering)
      ring_mol = mol.get_new_induced_subgraph( vring, ering)
      # here we need to take care of aromatic bonds
      # this is needed to properly match naphthalene, etc.
      if detect_aromatic:
        for e in ring_mol.edges:
          if e.aromatic:
            e.order = 4
        ring_mol.localize_aromatic_bonds()
      ring_hash = ring_mol.get_structure_hash()
      if ring_hash in self.rings:
        hit = ring_match( vring, self.rings[ ring_hash])
      else:
        ring_obj = ring( None, smiles.mol_to_text( ring_mol), ring_hash=ring_hash)
        self.rings[ ring_hash] = ring_obj
        hit = ring_match( vring, ring_obj)
      hits.append( hit)
    return hits

  @classmethod
  def _data_files_to_python_module( self, structure_file=None, ring_file=None):
    with open("subsearch_data.py", 'w') as out:
      print("""#--------------------------------------------------------------------------
#     This file is part of OASA - a free chemical python library
#     Copyright (C) 2003-2008 Beda Kosata <beda@zirael.org>

#     This program is free software; you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation; either version 2 of the License, or
#     (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.

#     Complete text of GNU GPL can be found in the file gpl.txt in the
#     main directory of the program

#--------------------------------------------------------------------------

""", file=out)
      print("## automatically generated file - may be overwritten at any time", file=out)
      print("structures = [", file=out)

      with open(structure_file or self.substructure_def_file, 'r') as f:
        for line in f:
          l = line.strip()
          if l and not l.startswith("#"):
            parts = [x.strip() for x in l.split(";")]
            if len(parts) < 3:
              print("Invalid line in src file:", line[:-1], file=sys.stderr)
            elif len(parts) == 3:
              parts.append("")
            to_ignore = map(int, filter(None, parts[3].split(",")))
            parts[3] = to_ignore
            if not parts[1]:
              parts[1] = parts[0]
            print(tuple(parts), ",", file=out)
        print("]", file=out)

      with open(ring_file or self.ring_def_file, 'r') as f:
        print("rings = [", file=out)
        for line in f:
          l = line.strip()
          if l and not l.startswith("#"):
            parts = [x.strip() for x in l.split(";")]
            print(tuple(parts), ",", file=out)
        print("]", file=out)


class substructure( object):

  def __init__( self, name, compound_type, smiles="", atoms_to_ignore=None):
    self.name = name
    self.compound_type = compound_type
    self.structure = None
    if smiles:
      self.read_smiles( smiles, atoms_to_ignore=atoms_to_ignore)
    self.children = []

  def __str__( self):
    return "<Substructure: %s: %s>" % (self.name, self.smiles_string)

  def read_smiles( self, smiles_string, atoms_to_ignore=None):
    self.smiles_string = smiles_string.strip()
    self.structure = smiles.text_to_mol( smiles_string, calc_coords=False)
    if atoms_to_ignore:
      self.atoms_to_ignore = [self.structure.vertices[x-1] for x in atoms_to_ignore]
    else:
      self.atoms_to_ignore = []

  def find_matches( self, mol):
    ret = []
    ms = list( mol.select_matching_substructures( self.structure, implicit_freesites=True, auto_cleanup=False))
    for atoms in ms:
      num = min( atoms[0].properties_['subsearch'].keys())
      atoms_in_fragment = [a.properties_['subsearch'][num] for a in atoms]
      ret.append( substructure_match( atoms, atoms_in_fragment, self))
    mol.clean_after_search( self.structure)
    return ret


class ring( object):

  def __init__( self, name, smiles, ring_hash=None):
    self.name = name
    self.compound_type = "ring"
    self.smiles_string = smiles.strip()
    if ring_hash:
      self.ring_hash = ring_hash
    else:
      self.ring_hash = None # read smiles and generate hash here?

  def __str__( self):
    return "<Ring: %s: %s>" % (self.name, self.smiles_string)


class substructure_match( object):

  def __init__( self, atoms_found, atoms_searched, substruct):
    """atoms_found are atoms in the molecule we searched in,
    atoms_searched are atoms in the fragment we used for search,
     (atoms_searched and atoms_found are guaranteed to have the same order, thus
      allowing matching between the two structures),
    substruct is the substructure instance"""
    self.substructure = substruct
    self.atoms_found = atoms_found
    self.atoms_searched = atoms_searched

  def __str__( self):
    return "<Match of %s with %d atoms (%d significant)>" % (self.substructure, len( self.atoms_found), len( self.get_significant_atoms()))

  def get_significant_atoms( self):
    ret = []
    for i,af in enumerate( self.atoms_found):
      atms = self.atoms_searched[i]
      if atms not in self.substructure.atoms_to_ignore:
        ret.append( af)
    return ret


class ring_match( object):

  def __init__( self, atoms_found, ring_obj):
    self.substructure = ring_obj
    self.atoms_found = atoms_found

  def __str__( self):
    return "<RingMatch of %s>" % (self.substructure)

  def get_significant_atoms( self):
    return self.atoms_found



if __name__ == "__main__":
  # update the subsearch_data.py module
  substructure_search_manager._data_files_to_python_module()
  import time
  t = time.time()
  ssm = substructure_search_manager()
  #ssm._read_structure_file()
  #ssm._read_ring_file()

  print("Read_structure_file: %.1fms" % (1000*(time.time() - t)))
  t = time.time()

  print(ssm.structures)
  #print(ssm.structures.is_connected())
  dump = ssm.structures.get_graphviz_text_dump()
  with open("dump.dot", 'w') as f:
    f.write(dump)

  print("Graphviz dump: %.1fms" % (1000*(time.time() - t)))
  t = time.time()

  def print_tree( x, l):
    print(l*" ", x) #, x.children
    for ch in x.children:
      print_tree( ch, l+2)

  #for tree in ssm._compute_search_trees():
  #  print_tree( tree, 0)

  #"c1cccc2c1cccc2
  #text = "COc5ccc4c2sc(cc2nc4c5)-c(cc1nc3c6)sc1c3ccc6OC"
  text = 'C(=O)OCC'
  #text2 = 'C(=O)[O-]'
  mol = smiles.text_to_mol( text, calc_coords=False)
  #m2 = smiles.text_to_mol( text2, calc_coords=False)
  #print("XXXX", mol.contains_substructure( m2))

  subs = ssm.find_substructures_in_mol( mol)
  for sub in subs:
    print(sub)

  print("Substructure search: %.1fms" % (1000*(time.time() - t)))
  t = time.time()



#--------------------------------------------------------------------------
#     This file is part of OASA - a free chemical python library
#     Copyright (C) 2003-2008 Beda Kosata <beda@zirael.org>

#     This program is free software; you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation; either version 2 of the License, or
#     (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.

#     Complete text of GNU GPL can be found in the file gpl.txt in the
#     main directory of the program

#--------------------------------------------------------------------------

from .molecule import molecule



class reaction(object):
  """Reaction representation.

  """
  def __init__( self, reactants=None, products=None, reagents=None):
    self.reactants = reactants or []
    self.products = products or []
    self.reagents = reagents or []


  def __str__(self):
    s = "\nreactants:\n"
    for i in self.reactants:
      s += "    {0}\n".format(i)
    s += "reagents:\n"
    for i in self.reagents:
      s += "    {0}\n".format(i)
    s += "products:\n"
    for i in self.products:
      s += "    {0}\n".format(i)

    return s



class reaction_component(object):
  """Represents one component of a reaction.

  """
  def __init__( self, mol=None, stoichiometry=1):
    self.stoichiometry = stoichiometry
    self.molecule = mol


  @property
  def molecule(self):
    return self._molecule


  @molecule.setter
  def molecule(self, mol):
    assert isinstance(mol, molecule)
    self._molecule = mol


  @property
  def stoichiometry(self):
    return self._stoichiometry


  @stoichiometry.setter
  def stoichiometry(self, stoich):
    assert isinstance(stoich, (int, float))
    self._stoichiometry = stoich


  def __str__( self):
    return "%s * (%s)" % (self.stoichiometry, self.molecule)




#--------------------------------------------------------------------------
#     This file is part of OASA - a free chemical python library
#     Copyright (C) 2003-2008 Beda Kosata <beda@zirael.org>

#     This program is free software; you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation; either version 2 of the License, or
#     (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.

#     Complete text of GNU GPL can be found in the file gpl.txt in the
#     main directory of the program

#--------------------------------------------------------------------------

from __future__ import print_function

import re
import os
import sys
import string
import select
import xml.dom.minidom as dom

from . import misc
from . import molfile
from . import dom_extensions
from . import coords_generator
from . import periodic_table as pt
from .plugin import plugin
from .molecule import molecule
from .stereochemistry import cis_trans_stereochemistry
from .oasa_exceptions import oasa_not_implemented_error, oasa_inchi_error, oasa_unsupported_inchi_version_error



class inchi( plugin):

  name = "inchi"
  read = 1
  write = 0

  proton_donors = ['O','F','Cl','Br','I','N','S']
  proton_acceptors = ['P','N','S','O']
  electron_donors = ['O','N','S','P']
  electron_acceptors = ['B']

  stereo_remap = {"+": cis_trans_stereochemistry.OPPOSITE_SIDE,
                  "-": cis_trans_stereochemistry.SAME_SIDE,
                  "?": cis_trans_stereochemistry.UNDEFINED,
                  "u": cis_trans_stereochemistry.UNDEFINED,
                  }


  def __init__( self, structure=None):
    self.structure = structure
    self._no_possibility_to_improve = False
    self.cleanup()

  def cleanup( self):
    self.charge = 0
    self.forced_charge = 0
    self._protonation_dealt_with_already = 0
    self._added_hs = set()

  def set_structure( self, structure):
    self.structure = structure

  def get_structure( self):
    return self.structure



  def split_layers( self, text):
    # try if we have a xml document in hand
    try:
      doc = dom.parseString( text)
    except:
      # it seems not to be the case
      if text.startswith( "InChI="):
        text = text[6:]
      layers = re.split( "/", text.strip())
      # fixed hydrogen layer is marked stupidly by f/h :(, we have to deal with it
      i = None
      for l in layers:
        if l.startswith("f"):
          i = layers.index( l)
          break
      if i is not None:
        del layers[i]
        layers[i] = 'f'+layers[i][1:]
      return layers
    structs = doc.getElementsByTagName( 'structure')
    if not structs:
      raise Exception("no structures found in xml string %s" % text)
    struct = structs[0]
    layers = []
    for name in ('version','formula','connections','H'):
      try:
        layer = struct.getElementsByTagName( name)[0]
      except IndexError:
        raise Exception("no '%s' tag found in xml string" % name)
      layers.append( dom_extensions.getTextFromElement( layer))
    return layers



  def get_layer( self, prefix):
    for l in self.layers:
      if l.startswith( prefix):
        return l[1:]

  def get_atom_with_inchi_number( self, number):
    assert type( number) == int
    for v in self.structure.vertices:
      if v.properties_['inchi_number'] == number:
        return v

  def read_inchi( self, text):
    try:
      self._read_inchi( text)
    except AssertionError:
      raise oasa_inchi_error( "Localization of bonds, charges or movable hydrogens failed")
    except:
      raise


  def _read_inchi( self, text):
    if not text:
      raise oasa_inchi_error( "No inchi was given")
    self.structure = molecule()
    self.layers = self.split_layers( text)
    # version support (very crude)
    self.version = self._check_version( self.layers[0])
    if not self.version:
      raise oasa_unsupported_inchi_version_error( self.layers[0])
    elif str( self.version[0]) != '1' or str( self.version[1]) != '0':
      raise oasa_unsupported_inchi_version_error( self.layers[0])

    self.hs_in_hydrogen_layer = self.get_number_of_hydrogens_in_hydrogen_layer()
    self.read_sum_layer()
    self.read_connectivity_layer()
    self._charge_mover = self._move_charge_somewhere_else()
    repeat = True
    run = 0
    # we have to repeat this step in order to find the right positioning of movable hydrogens
    while repeat and not self._no_possibility_to_improve:
      # cleanup
      for h in self._added_hs:
        self.structure.remove_vertex( h)
      for b in self.structure.edges:
        b.order = 1
      for v in self.structure.vertices:
        v.symbol = v.symbol
        v.charge = 0
      self.cleanup()

      # the code itself
      run += 1
      assert run < 50
      self._deal_with_notorious_groups()
      self.process_forced_charges()
      self.read_hydrogen_layer( run=run)
      self.read_charge_layer()
      self.read_p_layer()
      self.deal_with_da_bonds()
      #self._deal_with_valencies()
      self.compensate_for_forced_charges()
      self.structure.add_missing_bond_orders()
      #self.read_double_bond_stereo_layer()

      # here we check out if the molecule seems ok
      fvs = [v for v in self.structure.vertices if v.free_valency]
      if not fvs and not list(filter(None, [not v.order for v in self.structure.edges])):
        repeat = False
      else:
        if len( fvs) == 1:
          a = fvs[0]
          a.symbol = a.symbol # this resets the valency
          a.raise_valency_to_senseful_value()
          if not a.free_valency:
            repeat = False
          # deal with CO and other that require adding both + and -
          elif a.free_valency == 2 and [n for n in a.neighbors if n.free_valency==0 and n.symbol in self.electron_donors]:
            neighbor = [n for n in a.neighbors if n.free_valency==0 and n.symbol in self.electron_donors][0]
            a.charge -= 1
            neighbor.charge += 1
            a.get_edge_leading_to( neighbor).order = 1
            self.structure.add_missing_bond_orders()
            if not [v for v in self.structure.vertices if v.free_valency > 0]:
              # we fixed it
              repeat = False


      if repeat and self._no_possibility_to_improve and self.charge:
        try:
          next(self._charge_mover)
        except StopIteration:
          pass
        else:
          self._no_possibility_to_improve = False
          run = 0

    if repeat and self._no_possibility_to_improve:
##       if len( filter( None, [v.free_valency for v in self.structure.vertices])) == 1:
##         print
##         print([(v.symbol, v.valency, v.free_valency)  for v in self.structure.vertices if v.free_valency], filter( None, [not v.order for v in self.structure.edges]), text)
##       if sum( [v.charge for v in self.structure.vertices]) != self.charge:
##         print("Charge problem", sum( [v.charge for v in self.structure.vertices]), self.charge)
      #pass
      raise oasa_inchi_error( "Localization of bonds, charges or movable hydrogens failed")
#    print("run:", run, file=sys.stderr)



  def read_sum_layer( self):
    if "." in self.layers[1]:
      raise oasa_not_implemented_error( "INChI", "multiple compound systems are not supported by the library")

    form = pt.formula_dict( self.layers[1])
    processed_hs = 0 #for diborane and similar compounds we must process some Hs here
    j = 0
    for k in form.sorted_keys():
      for i in range( form[k]):
        if k == 'H':
          # we want to process only the Hs that are not in the h-layer
          if processed_hs >= form[k] - self.hs_in_hydrogen_layer:
            continue
          else:
            processed_hs += 1
        j += 1
        a = self.structure.create_vertex()
        a.symbol = k
        self.structure.add_vertex( a)
        a.properties_['inchi_number'] = j



  def read_connectivity_layer( self):
    layer = self.get_layer( "c")
    if not layer:
      return
    chunks = re.split( "([0-9]*)", layer)
    chunks = filter( None, chunks)
    chunks = filter( lambda x: x!='-', chunks)
    last_atom = None
    bracket_openings = []
    for c in chunks:
      if c == '(':
        bracket_openings.append( last_atom)
      elif c == ')':
        last_atom = bracket_openings.pop(-1)
      elif c == ",":
        last_atom = bracket_openings.pop(-1)
        bracket_openings.append( last_atom)
      else:
        try:
          i = int( c)
        except:
          raise ValueError("unexpected character %s in the connectivity layer" % c)
        # atom
        if last_atom:
          self.structure.add_edge( last_atom-1, i-1)
        last_atom = i


  def read_hydrogen_layer( self, run=0):
    layer = self.get_layer( "h")
    # check presence of the layer
    if not layer:
      return

    fixed_layer = self.get_layer( "f")
    if fixed_layer:
      # we can skip movable hydrogens
      self._read_simple_hydrogen_layer( fixed_layer)
    else:
      re_for_brackets = "\([H\d,\-]+?\)"
      brackets = re.findall( re_for_brackets, layer)
      for bracket in brackets:
        self._process_moving_hydrogen( bracket[1:-1], run=run)

    # we improve only in the _process_moving_hydrogen so if it was not called there is no possibility for improvement
    if not self._added_hs or fixed_layer:
      self._no_possibility_to_improve = True

    self._read_simple_hydrogen_layer( layer)



  def _read_simple_hydrogen_layer( self, layer):
    """just takes the layer and adds hydrogens according to what it seas,
    it does not care about moving hydrogens and stuff"""

    re_for_brackets = "\([H\d,\-]+?\)"
    layer = re.sub( re_for_brackets, "", layer)  # clean the brackets out

    for vs, num in self._parse_h_layer( layer):
      for v in vs:
        for j in range( num):
          h = self.structure.create_vertex()
          h.symbol = 'H'
          self.structure.add_vertex( h)
          self.structure.add_edge( v-1, h)
          self._added_hs.add( h)



  def read_charge_layer( self):
    layer = self.get_layer( "q")
    if not layer:
      return

    charge = int( layer[1:]) - self.forced_charge
    self.charge = int( layer[1:])


    if charge > 0:
      # at first we try to put positive charge on atoms with negative free_valency
      change = True
      steps = 0
      while charge and change:
        steps += 1
        assert steps < 10
        change = False
        for v in self.structure.vertices:
          if v.free_valency == -1:
            if charge > 0:
              v.charge = -v.free_valency
            else:
              v.charge = v.free_valency
            charge -= v.charge
            change = True
            break

    if charge < 0:
      # we try to put negative charge on orphan atoms
      steps = 0
      change = True
      while charge and change:
        steps += 1
        assert steps < 10
        change = False
        for v in self.structure.vertices:
          if v.free_valency and sum( [n.free_valency for n in v.neighbors]) < v.free_valency:
            charge = self._valency_to_charge( v, charge)
            change = True
            break

    # this could be rewritten to put charges to places where a segment with odd number of free_valencies is
    # this would be more general than counting the free_valencies for the whole molecule
    free_valencies = sum( [a.free_valency for a in self.structure.vertices])
    if free_valencies % 2:
      # odd number of free_valencies means we have to put the charge somewhere to make a free_valency there
      change = True
      steps = 0

      while charge and change:
        change = False
        steps += 1
        assert steps < 10

        for v in self.structure.vertices:
          if v.symbol != 'C' and not v.free_valency and list(filter(None, [n.free_valency for n in v.neighbors])):
            v.charge = charge
            charge = 0
            change = True
            break

    # then we do the others, at first trying heteroatoms
    steps = 0
    change = True
    while charge and change:
      change = False
      steps += 1
      assert steps < 10

      for v in self.structure.vertices:
        if v.free_valency and v.symbol != 'C':
          charge = self._valency_to_charge( v, charge)
          change = True
          break


    # then we try the carbon atoms as well
    steps = 0
    change = True
    while charge and change:
      steps += 1
      assert steps < 10

      change = False
      for v in self.structure.vertices:
        if v.free_valency:
          charge = self._valency_to_charge( v, charge)
          change = True
          break

    if charge:
      raise oasa_exceptions.oasa_inchi_error( "The molecular charge could not be allocated to any atom (%d)." % charge)



  def read_p_layer( self):
    layer = self.get_layer( "p")
    if not layer:
      return


    p = int( layer[1:]) - self._protonation_dealt_with_already
    self.charge += p
    charges = self.forced_charge
    old_p = p

    # at first add protons to negative charges
    for v in self.structure.vertices:
      if p > 0:
        if v.charge < 0:
          h = self.structure.create_vertex()
          h.symbol = 'H'
          self.structure.add_vertex( h)
          self.structure.add_edge( v, h)
          p -= 1
          v.charge += 1
          self._protonation_dealt_with_already += 1

    if self._protonation_dealt_with_already:
      # p was solved above
      dp = self._protonation_dealt_with_already - sum( [v.charge for v in self.structure.vertices])
      if dp > 0:
        # there were no forced charges, so we are not in a zwittrion state, huh, who does not understand that?!
        for v in self.structure.vertices:
          if dp > 0 and v.symbol in self.proton_acceptors:
            v.charge += 1
            dp -= 1

    # then come the other additional protones
    while p:
      old_p = p
      for v in self.structure.vertices:
        if p > 0:
          if v.symbol in self.proton_acceptors:
            h = self.structure.create_vertex()
            h.symbol = 'H'
            self.structure.add_vertex( h)
            self.structure.add_edge( v, h)
            self._added_hs.add( h)
            p -= 1
            if charges:
              # the charge is already present as a forced charge
              charges -= 1
            else:
              # the charge can normaly be delt with
              v.charge += 1
            break
        elif p < 0:
          if v.symbol in self.proton_donors:
            hs = [n for n in v.neighbors if n.symbol=='H']
            if hs:
              h = hs[0]
              self.structure.remove_vertex( h)
              p += 1
              if charges:
                # the charge is already present as a forced charge
                charges += 1
              else:
                # the charge can normaly be delt with
                v.charge -= 1
              break
      assert p < old_p


  def read_double_bond_stereo_layer( self):
    def get_lowest_numbered_neighbor( atom, verbotten_atom):
      neighs = [(n.properties_.get('inchi_number',1000000), n)
                for n in atom.neighbors
                if n is not verbotten_atom and "inchi_number" in n.properties_]
      if neighs:
        neighs.sort( reverse=True)
        return neighs[0][1]
      else:
        # we take the only one that is there (if any)
        neighs = [n for n in atom.neighbors if n is not verbotten_atom]
        if neighs:
          return neighs[0]
        else:
          raise oasa_inchi_error( "No neigbors on atom with stereo information!")



    layer = self.get_layer( "b")
    if not layer:
      return
    for a1,a2,sign in re.findall( "(\d+)-(\d+)([-+?u])", layer):
      atom1 = self.get_atom_with_inchi_number( int( a1))
      atom2 = self.get_atom_with_inchi_number( int( a2))
      bond = self.structure.get_edge_between( atom1, atom2)
      # we need neighbors with lowest inchi_number
      neigh1 = get_lowest_numbered_neighbor( atom1, atom2)
      neigh2 = get_lowest_numbered_neighbor( atom2, atom1)
      st = cis_trans_stereochemistry( references=[neigh1,atom1,atom2,neigh2],
                                      value=self.stereo_remap[sign])
      self.structure.add_stereochemistry( st)



  def _valency_to_charge( self, v, charge):
    """sets charge of the atom so that it has minimal free_valency,
    returns the 'unused' charge in case it is not comsumed whole"""
    ch = misc.signum( charge) * min( abs( charge), v.free_valency)
    if charge < 0:
      charge += ch
      v.charge = -ch
    else:
      charge -= ch
      v.charge = ch
    return charge


  def get_number_of_hydrogens_in_hydrogen_layer( self):
    # version check
    layer = self.get_layer( "h")
    if not layer:
      return 0

    # check if we can handle it
    if "*" in layer or ";" in layer:
      raise oasa_not_implemented_error( "INChI", "multiple compound systems are not supported by the library")

    ret = 0

    re_for_brackets = "\([H\d,\-]+?\)"
    brackets = re.findall( re_for_brackets, layer)
    for bracket in brackets:
      ret += self._get_hs_in_moving_hydrogen( bracket[1:-1])
    layer = re.sub( re_for_brackets, "", layer)  # clean the brackets out

    for vs, num in self._parse_h_layer( layer):
      ret += len( vs) * num

    return ret


  def _get_hs_in_moving_hydrogen( self, chunk):
    chks = chunk.split( ',')
    if len( chks[0]) > 1:
      if chks[0][1:].endswith( "-"):
        hs = int( chks[0][1:-1] or 1)  # number of atoms
      else:
        hs = int( chks[0][1:])  # number of atoms
    else:
      hs = 1
    return hs



  def _check_version( self, ver):
    """returns a tuple of two version numbers"""
    if "Beta" in ver:
      return None
    v = ver.strip(string.ascii_lowercase+string.ascii_uppercase)
    if "." in v:
      return v.split('.')
    else:
      return v, 0


  def _process_moving_hydrogen( self, chunk, run=0):
    chks = chunk.split( ',')
    if len( chks[0]) > 1:
      if chks[0][1:].endswith("-"):
        hs = int( chks[0][1:-1] or 1) + 1  # number of atoms
        self._protonation_dealt_with_already += 1
      else:
        hs = int( chks[0][1:])  # number of atoms
    else:
      hs = 1

    vertices = [self.structure.vertices[ int( i)-1] for i in chks[1:]]
    vs = vertices
    take = hs

    # here we generate still shorter combinations of vertices that receive hydrogens
    # because of applying them in circles in the next step, we simulate the case
    # where more than one hydrogen is put to one atom
    while (run > misc.x_over_y( len( vs), take)) and take:
      run -= misc.x_over_y( len( vs), take)
      take -= 1
    if not take:
      self._no_possibility_to_improve = True
      return []
    variations = misc.gen_variations( vs, take)
    for i in range( run):
      vs = next(variations)

    while hs:
      for v in vs:
        h = self.structure.create_vertex()
        h.symbol = 'H'
        self.structure.add_vertex( h)
        self.structure.add_edge( v, h)
        self._added_hs.add( h)
        hs -= 1
        if not hs:
          break



  def _split_h_layer( self, layer):
    was_h = False
    chunks = []
    chunk = ""
    for c in layer:
      if c == 'H':
        was_h = True
        chunk += c
      elif c == "," and was_h:
        was_h = False
        chunks.append( chunk)
        chunk = ""
      else:
        chunk += c
    if chunk:
      chunks.append( chunk)
    return chunks


  def _parse_h_layer( self, layer):
    chunks = self._split_h_layer( layer)
    for chunk in chunks:
      try:
        head, tail = chunk.split( 'H')
      except Exception as e:
        raise oasa_inchi_error( "error in hydrogen layer - missing H symbol")
      num_h = tail and int( tail) or 1
      vertices = []
      for p in head.split( ","):
        if "-" in p:
          try:
            a, b = map( int, p.split("-"))
          except Exception as e:
            raise oasa_inchi_error( "error in hydrogen layer - non-number character(s) present in atom range specification")
          vertices.extend( range( a, b+1))
        else:
          try:
            vertices.append( int( p))
          except Exception as e:
            raise oasa_inchi_error( "error in hydrogen layer - non-number character(s) present in atom specification")

      yield vertices, num_h


  def process_forced_charges( self):
    """this marks the charges that are forced by the connectivity and thus helps
    process zwitrions"""
    forced_charge = 0
    for v in self.structure.vertices:
      if v.symbol in ("N","S") and v.free_valency == -1:
        v.charge = 1
        forced_charge += 1
    self.forced_charge = forced_charge


  def compensate_for_forced_charges( self):
    """if there were foced charges and the molecule should not have any charge,
    we have to take care of it here"""
    charge = self.charge - self.forced_charge
    old_charge = charge
    while not self.charge and charge:
      for chunk in self.structure._gen_free_valency_connected_components():
        if len( chunk) % 2:
          if len( chunk) == 1:
            vs = chunk
          else:
            vs = [v for v in chunk if len( set( v.neighbors) & chunk) == 1]
          for v in vs:
            if charge < 0:
              if v.symbol in self.proton_donors and v.free_valency:
                v.charge -= 1
                charge += 1
            elif charge > 0:
              if v.symbol in self.proton_acceptors:
                v.charge += 1
                charge -= 1
            if not charge:
              return
##       if old_charge == charge:
##         print("AAAAA", self.layers)
      assert old_charge != charge


  def deal_with_da_bonds( self):
    """deal with donor-acceptor bonds, this fixes mostly boron containing compounds"""
    for v in self.structure.vertices:
      if v.free_valency < 0 and v.symbol in self.electron_acceptors:
        ns = [n for n in v.neighbors if n.symbol in self.electron_donors]
        if ns:
          n = ns[0]
          n.charge += 1
          v.charge -= 1

  def _move_charge_somewhere_else( self):
    if self.charge < 0:
      pass
    elif self.charge == 1:
      for v in self.structure.vertices:
        if v.symbol in self.proton_acceptors:
          v.charge = 1
          yield None


  def _deal_with_notorious_groups( self):
    """some groups such as NO2, SO3H etc. need the valency of the central atom to be risen,
    this is done here"""
    for v in self.structure.vertices:
      if v.symbol == "N" and v.degree > 2:
        for n in v.neighbors:
          if n.symbol == "O" and n.degree == 1:
            v.raise_valency()
            break
      elif v.symbol == "S" and v.symbol > 2:
        for n in v.neighbors:
          if n.symbol == "O" and n.degree == 1:
            v.raise_valency()


  def _deal_with_valencies( self):
    go = True
    i = 0
    while go:
      i += 1
      if i> 100:
        print("aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa")
        break
      go = False
      # set all valency_raise to zero
      for v in self.structure.vertices:
        v.properties_['valency_raise'] = 0
      # mark vertices with the demands to raise valency
      for comp in self.structure._gen_free_valency_connected_components():
#        print(map( str, comp))
        if sum( [v.free_valency for v in comp]) % 2:
          go = True
          for v in comp:
            for ne in v.neighbors:
              if ne not in comp:
                ne.properties_['valency_raise'] += 1
      # raise the valency of the most asked for atom
      vs = sorted([(v.properties_['valency_raise'], v) for v in self.structure.vertices],
                  reverse=True)
      for (x, v) in vs:
#        print(x, v, v.valency, v.free_valency)
        if x and v.raise_valency():
          break
##           if v.free_valency == 0 and v.symbol in self.proton_acceptors:
##             if v.charge == 0:
##               v.charge = 1
##             elif v.charge == 1:
##               if v.raise_valency():
##                 v.charge = 0
##                 break
##           elif v.raise_valency():
##             break
#          print(x, v, v.symbol, v.degree, v.valency, v.free_valency)
#          print([e.order for e in v.neighbor_edges])
#          break
#        elif x == 0:
#          break

import subprocess

def _run_command(command, inputs):
  p = subprocess.Popen(command,
                       stdin=subprocess.PIPE,
                       stdout=subprocess.PIPE,
                       stderr=subprocess.STDOUT)
  o, er = p.communicate(inputs.encode('utf-8'))
  return o.decode('utf-8')


def generate_inchi_and_inchikey( m, program=None, fixed_hs=True, ignore_key_error=False):
  """ignore the case when InChIKey cannot be generated for some reason
  (no mhash library and old InChI program)"""
  if not program:
    from . import config
    program = config.Config.inchi_binary_path
  mf = molfile.mol_to_text( m)
  if os.name == 'nt':
    options = "/AUXNONE /STDIO /Key" + (fixed_hs and " /FixedH" or "")
  else:
    options = "-AUXNONE -STDIO -Key" + (fixed_hs and " -FixedH" or "")
  #print(options)
  command = [os.path.abspath( program)] + options.split()
  text = _run_command( command, mf)
  inchi = ""
  key = ""
  warnings = []
  for line in text.splitlines():
    if line.startswith( "Warning"):
      warnings.append(line.strip() + "\n")
    elif line.startswith( "End of file detected"):
      pass
    elif line.startswith( "InChIKey="):
      key = line.strip()[9:]
      break
    elif line.startswith( "InChI="):
      inchi = inchi + line.strip()
    elif line.startswith( "Error"):
      break
  if not inchi:
    raise oasa_inchi_error( "InChI program did not create any output InChI")
  if not key:
    # probably old version of the InChI software
    try:
      from . import inchi_key
    except ImportError:
      if ignore_key_error:
        key = None
      else:
        raise oasa_inchi_error( "InChIKey could not be generated - inchi_key module failed to load properly.")
    else:
      key = inchi_key.key_from_inchi( inchi)
  return inchi, key, warnings


def generate_inchi( m, program=None, fixed_hs=True):
  inchi, key, warnings = generate_inchi_and_inchikey( m, program=program, fixed_hs=fixed_hs, ignore_key_error=True)
  return inchi, warnings

def generate_inchi_key( m, program=None, fixed_hs=True):
  inchi, key, warnings = generate_inchi_and_inchikey( m, program=program, fixed_hs=fixed_hs)
  return key, warnings



##################################################
# MODULE INTERFACE

reads_text = 1
reads_files = 1
writes_text = 1
writes_files = 1

def text_to_mol( text, include_hydrogens=True, mark_aromatic_bonds=False, calc_coords=1):
  try:
    inc = inchi()
    inc.read_inchi( text)
    mol = inc.structure
    #mol.add_missing_bond_orders()
    if not include_hydrogens:
      mol.remove_unimportant_hydrogens()
    if calc_coords:
      coords_generator.calculate_coords( mol, bond_length=calc_coords)
    if mark_aromatic_bonds:
      mol.mark_aromatic_bonds()
    return mol
  except oasa_inchi_error:
    raise
  except oasa_not_implemented_error:
    raise
  except:
    raise

def file_to_mol( f):
  return text_to_mol( f.read())


def mol_to_text( mol, program=None, fixed_hs=True):
  return generate_inchi( mol, program=program, fixed_hs=fixed_hs)[0]


def mol_to_file( mol, f):
  f.write( mol_to_text( mol))



#
##################################################




##################################################
# DEMO

if __name__ == '__main__':
  import time

  from . import smiles

  def main( text, cycles):
    t1 = time.time()
    for jj in range( cycles):
      mol = text_to_mol( text, calc_coords=True, include_hydrogens=False)
      print(map( str, [b for b in mol.bonds if b.order == 0]))
      print("  smiles: ", smiles.mol_to_text( mol))
      print("  inchi:  ", generate_inchi( mol, fixed_hs=False, program="/home/beda/bin/stdinchi-1"))
      print("  charge: ", sum( [a.charge for a in mol.vertices]))
      print("  mw:     ", mol.weight)
    print(generate_inchi_and_inchikey( mol, fixed_hs=False, program="/home/beda/bin/stdinchi-1"))
    t1 = time.time() - t1
    print('time per cycle', round( 1000*t1/cycles, 2), 'ms')

  repeat = 3
  inch = "InChI=1S/C34H16O2/c35-33-25-7-3-1-5-17(25)19-9-11-21-24-14-16-28-32-20(18-6-2-4-8-26(18)34(28)36)10-12-22(30(24)32)23-13-15-27(33)31(19)29(21)23/h1-16H"

  #"InChI=1/C34H16O2/c35-33-25-7-3-1-5-17(25)19-9-11-21-24-14-16-28-32-20(18-6-2-4-8-26(18)34(28)36)10-12-22(30(24)32)23-13-15-27(33)31(19)29(21)23/h1-16H"
  #inch = "InChI=1S/CO/c1-2"
  print("oasa::INCHI DEMO")
  print("converting following inchi into smiles (%d times)" % repeat)
  print("  inchi:   %s" % inch)

  #import profile
  #profile.run( 'main( inch, repeat)')
  main( inch, repeat)


# DEMO END
##################################################



##################################################
## TODO

## rename layer to sublayer


##################################################




#--------------------------------------------------------------------------
#     This file is part of OASA - a free chemical python library
#     Copyright (C) 2003-2008 Beda Kosata <beda@zirael.org>

#     This program is free software; you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation; either version 2 of the License, or
#     (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.

#     Complete text of GNU GPL can be found in the file gpl.txt in the
#     main directory of the program

#--------------------------------------------------------------------------

"""Support module for some geometric mesurements.

(Geometric tramforms are in transform.py.)
"""

from __future__ import division

import operator

from math import sqrt, atan2, pi, cos, sin

from .misc import signum, normalize_coords



def find_parallel( x1, y1, x2, y2, d):
  "returns tuple of coordinates for parallel abscissa in distance d"
  # following is here to ensure that signum of "d" clearly determines
  # the side of line on whitch the parallel is drawn
  if round( y2, 3) -round( y1, 3) != 0:
    if y2 < y1:
      d = -d
    k = -(x2-x1)/(y2-y1)
    x = ( d + sqrt( k**2 +1)*x1)/ sqrt( k**2 +1)
    y = y1 + k*( x -x1)
    x0 = ( d + sqrt( k**2 +1)*x2)/ sqrt( k**2 +1)
    y0 = y2 + k*( x0 -x2)
  else:
    if x1 > x2:
      d = -d
    x, x0 = x1, x2
    y = y1 - d
    y0 = y2 - d
  return (x, y, x0, y0)


def get_parallel_distance( l1, l2):
  "returns distance of two parallels - does not check whether the two are really parallels"
  x1a, y1a, x2a, y2a = l1
  x1b, y1b, x2b, y2b = l2
  if x1a == x2a:
    return y1a-y2a
  elif y1a == y2a:
    return x1a-x2a
  else:
    ka = (y2a - y1a)/(x2a - x1a)
    qa = y1a - ka*x1a
    kb = ka
    qb = y1b - kb*x1b
    k_rev = -1/ka
    q_rev = y1a - k_rev*x1a
    x0 = (qb - q_rev)/(k_rev - kb)
    y0 = k_rev*x0 + q_rev
    d = sqrt( (x1a-x0)**2 + (y1a-y0)**2)
    if qa < qb:
      d = -d
    return d


def get_parallel_signum( l1, l2):
  x1a, y1a, x2a, y2a = l1
  x1b, y1b, x2b, y2b = l2
  if x1a == x2a:
    return signum( -x1a + x1b)
  elif y1a == y2a:
    return signum( -y1a+y1b)
  else:
    return signum( -x2a+x2b)


def on_which_side_is_point( line, point, threshold=0):
  """tells whether a point is on one side of a line or on the other (1,0,-1) - 0 is for point on line.
  line is given as sequence of four coordinates, point as sequence of two coords,
  threshold means what smallest angle is considered to still be on the line"""
  x1, y1, x2, y2 = line
  x, y = point
  a = atan2( y-y1, x-x1)
  b = atan2( y2-y1, x2-x1)
  # If a is too small, pick the other point
  if abs(a) <= threshold:
    a = atan2(y-y2, x-x2)
  if a*b < 0 and abs(a-b) > pi:
    if a < 0:
      a += 2*pi
    else:
      b += 2*pi
  if abs( a-b) <= threshold or abs( abs( a-b) -pi) <= threshold:
    return 0
  elif a-b < 0:
    return 1
  else:
    return -1


def point_on_circle( center_x, center_y, radius, direction = (), resolution = 15):
  """finds point on circle in direction of (dx, dy), optionaly rounds the angle
  according to resolution"""
  dx, dy = direction
  if resolution:
    angle = round( atan2( dy, dx)/(pi*resolution/180.0))*(pi*resolution/180.0)
  else:
    angle = atan2( dy, dx)
  x = center_x + round( cos( angle) *radius, 2)
  y = center_y + round( sin( angle) *radius, 2)
  return x,y


def angle_between_lines( l1, l2):
  """returns angle between two lines"""
  a = clockwise_angle_from_east( l1[2]-l1[0], l1[3]-l1[1]) - clockwise_angle_from_east( l2[2]-l2[0], l2[3]-l2[1])
  if a > pi:
    return a - 2*pi
  elif a < -pi:
    return a + 2*pi
  else:
    return a


def clockwise_angle_from_east( dx, dy):
  """returns the angle in clockwise direction between the center-east line and direction"""
  angle = atan2( dy, dx)
  if angle < 0:
    angle = 2*pi + angle
  return angle


def intersection_of_line_and_rect( line, rect, round_edges=0):
  """finds a point where a line and a rectangle intersect,
  both are given as lists of len == 4"""
  lx0, ly0, lx1, ly1 = map( float, line)
  rx0, ry0, rx1, ry1 = map( float, normalize_coords( rect))

  # find which end of line is in the rect and reverse the line if needed
  if (lx0 > rx0) and (lx0 < rx1) and (ly0 > ry0) and (ly0 < ry1):
    lx0, lx1 = lx1, lx0
    ly0, ly1 = ly1, ly0

  # the computation itself
  ldx = lx1 - lx0
  ldy = ly1 - ly0

  if abs( ldx) > 0.0001:
    # we calculate using y = f(x)
    k = ldy/ldx
    q = ly0 - k*lx0
    if ldx < 0:
      xx = rx1
    else:
      xx = rx0
    xy = k*xx + q
    # the result must be in the rectangle boundaries
    # but sometimes is not because rounding problems
    if not ry0 <= xy <= ry1:
      xx = lx0
      xy = ly0
  else:
    xx = lx0
    xy = ly0

  if abs( ldy) > 0.0001:
    # we calculate using x = f(y)
    k = ldx/ldy
    q = lx0 - k*ly0
    if ldy < 0:
      yy = ry1
    else:
      yy = ry0
    yx = k*yy + q
    # the result must be in the rectangle boundaries
    # but sometimes is not because rounding problems
    if not rx0 <= yx <= rx1:
      yy = ly0
      yx = lx0
  else:
    yy = ly0
    yx = lx0

  if point_distance( lx0, ly0, xx, xy) < point_distance( lx0, ly0, yx, yy):
    return (yx, yy)
  else:
    return (xx, xy)


def point_distance( x1, y1, x2, y2):
  return sqrt( (x2-x1)**2 + (y2-y1)**2)


def line_length(x1, y1, x2, y2):
  return point_distance( x1, y1, x2, y2)


def rectangle_intersection( rect1, rect2):
  pass


def do_rectangles_intersect( rect1, rect2):
  xs1 = [rect1[0], rect1[2]]
  xs2 = [rect2[0], rect2[2]]
  xs = xs1 + xs2
  ys1 = [rect1[1], rect1[3]]
  ys2 = [rect2[1], rect2[3]]
  ys = ys1 + ys2

  dx1 = abs( xs1[0] - xs1[1])
  dx2 = abs( xs2[0] - xs2[1])
  dx = max( xs) - min( xs)

  dy1 = abs( ys1[0] - ys1[1])
  dy2 = abs( ys2[0] - ys2[1])
  dy = max( ys) - min( ys)

  if dx1+dx2 > dx and dy1+dy2 > dy:
    return True
  else:
    return False


def is_point_inside_polygon( point, polygon):
  """point is a tuple or list of length == 2,
  polygon is a tuple or list of points"""
  a = 0
  for i in range( len( polygon)-1):
    a += angle_between_lines( point+polygon[i+1], point+polygon[i])
  a += angle_between_lines( point+polygon[0], point+polygon[ len( polygon)-1])
  a = abs( a)
  if a < 0.1:
    return False
  else:
    return True


def tkspline_to_quadratic_bezier( points):   #points = ((x1,y1),(x2,y2),...)
    if len(points) > 2:
      qbeziers = []
      startx,starty = points[0]             #start point
      curvex,curvey = points[1]             #point that defines curvature
      for nextx,nexty in points[2:-1]:
        endx = (curvex+nextx)/2             #calculate midway point
        endy = (curvey+nexty)/2             #to find end point
        qbeziers.append((startx,starty,curvex,curvey,endx,endy))

        startx,starty = endx,endy           #start where the last ended
        curvex,curvey = nextx,nexty

      endx,endy = points[-1]                #no midway calculation for last point
      qbeziers.append((startx,starty,curvex,curvey,endx,endy))

    else:
      startx,starty = points[0]             #If only two points are given
      endx,endy = points[1]                 #make a straight line
      curvex = (startx+endx)/2
      curvey = (starty+endy)/2
      qbeziers=((startx,starty,curvex,curvey,endx,endy),)

    return qbeziers


def tkspline_to_cubic_bezier( points):   #points = ((x1,y1),(x2,y2),...)
    qbeziers = tkspline_to_quadratic_bezier( points)
    cbeziers = []
    for qbez in qbeziers:
      startx, starty = qbez[0], qbez[1]
      endx, endy = qbez[4], qbez[5]
      qcurvex, qcurvey = qbez[2], qbez[3]

      ccurvexa, ccurveya = startx + 2/3*(qcurvex-startx), starty + 2/3*(qcurvey-starty)
      ccurvexb, ccurveyb = ccurvexa + 1/3*(endx-startx), ccurveya + 1/3*(endy-starty)
      cbeziers.append( (startx, starty, ccurvexa, ccurveya, ccurvexb, ccurveyb, endx, endy))
    # http://fontforge.sourceforge.net/bezier.html
    return cbeziers


def point_at_distance_from_line( x1, y1, x2, y2, d):
  "returns tuple of coordinates for a point in distance d from line orthogonal to point (x2,y2)"
  if round( y2, 3) -round( y1, 3) != 0:
    if y2 < y1:
      d = -d
    k = -(x2-x1)/(y2-y1)
    x0 = ( d + sqrt( k**2 +1)*x2)/ sqrt( k**2 +1)
    y0 = y2 + k*( x0 -x2)
  else:
    if x1 > x2:
      d = -d
    x0 = x2
    y0 = y2 - d
  return (x0, y0)


def intersection_of_two_lines (x1,y1,x2,y2,x3,y3,x4,y4,parallel_detection_threshold=3):
  """lines 1-2 and 3-4
  returns x,y, 0 if succesful or 1 if parallel
  y=mx+c is used
  parallel_detection_threshold is a negative decadic logarithm of minimal displacement of m
  that is considered parallel"""
  if x1-x2 == 0:
    if x3-x4 == 0:
      return 0,0,1,0        #lines paralell
    m2 = (y3-y4)/(x3-x4)
    c2 = y3 - m2 * x3
    rex,rey = x1,m2*x1+c2
  elif x3-x4 == 0:
    m1 = (y1-y2)/(x1-x2)
    c1 = y1 - m1 * x1
    rex,rey = x3,m1*x3+c1
  else:
    m1 = (y1-y2)/(x1-x2)
    m2 = (y3-y4)/(x3-x4)
    if round(m1-m2,parallel_detection_threshold) == 0:
      return 0,0,1,0            #lines paralell
    c2 = y3 - m2 * x3
    c1 = y1 - m1 * x1
    rex = -(c2-c1)/(m2-m1)
    rey = (c1*m2-c2*m1)/(m2-m1)
  #check if point is on the lines
  online1 = is_point_beween_points_of_line( (x1,y1,x2,y2), (rex,rey))
  online2 = is_point_beween_points_of_line( (x3,y3,x4,y4), (rex,rey))
  if online1 and online2:
    online = 3
  elif online1:
    online = 1
  elif online2:
    online = 2
  else:
    online = 0
  # x-coord, y-coord , paralell(0 or 1), on line (0=no line, 1=on line 1-2, 2=on line 3-4, 3=on both)
  return rex,rey,0,online


def mirror_point_on_line (xa,ya,x1,y1,x2,y2):
  """mirroring point a on line 1-2"""
  if x1-x2 == 0:                           #y=mc+c is used
    return 2*x1-xa,ya
  elif y1-y2 == 0:
    return xa,2*y1-ya
  else:
    m=(y1-y2)/(x1-x2)
    cl=y1-m*x1
    ca=ya+(1/m)*xa
    xbl = (ca-cl)/(m+1/m)   #point bl is intersection with line
    ybl = -(1/m)*xbl+ca
    xb = xbl*2-xa
    yb = ybl*2-ya
    return xb,yb


def elongate_line (x1,y1,x2,y2,d):
  """line 1-2 will be elongatet at point 2 negative d will make it shorter"""
  if x1-x2 == 0:
    rex = x2
    if y2 > y1:
      rey = y2 + d
    else:
      rey = y2 - d
  else:
    m = (y1-y2)/(x1-x2)
    dx = sqrt( d**2 / (1 + m**2))
    dy = m * dx
    if dy < 0:
      dy=-dy
    if d<0:
      dx,dy = -dx, -dy
    if x2 > x1:
      rex = x2 + dx
    else:
      rex = x2 - dx
    if y2 > y1:
      rey = y2 + dy
    else:
      rey = y2 - dy
  return rex, rey


def coordinate_flat_list_to_xy_tuples( coords):
  ret = []
  for i in range( 0, len( coords), 2):
    ret.append( (coords[i],coords[i+1]))
  return ret


def find_parallel_polyline( coords, d):
  """take and returns list of tuples of coordinates for parallel abscissa in distance d"""
  res = []
  i = 0
  for x,y in coords:
    if i == 0:
      x1, y1 = coords[1]
      res.append( point_at_distance_from_line( x1,y1,x,y,-d))
    elif i == len( coords)-1:
      x1, y1 = coords[-2]
      res.append( point_at_distance_from_line( x1,y1,x,y,d))
    else:
      #internal point
      x0, y0 = coords[i-1] #previous point
      x1, y1 = coords[i+1] #next point
      xa1,ya1,xa2,ya2 = find_parallel( x0, y0, x, y, d)
      xb1,yb1,xb2,yb2 = find_parallel( x1, y1, x, y, -d)
      xx, yy, parallel, online = intersection_of_two_lines( xa1,ya1,xa2,ya2,xb1,yb1,xb2,yb2,parallel_detection_threshold=1)
      if not parallel:
        if online:
          # it the cross-point is on one of the lines, this could not be the case
          # where the point runs too far from the original line
          res.append( (xx, yy))
        else:
          # we should check for very near parallels going back -
          # the cross-point is very far from the original line, we need to fix it
          xa, ya = elongate_line(xa1,ya1,xa2,ya2,abs(d))
          if point_distance(xa,ya,xa2,ya2) < point_distance(xa,ya,xx,yy):
            xb, yb = elongate_line(xb1,yb1,xb2,yb2,abs(d))
            res.append( (xa, ya))
            res.append( (xb, yb))
          else:
            res.append( (xx, yy))
      else:
        # is 1-2-3 straight or does 2-3 go back from 1-2
        if point_distance(x,y,x0,y0) > point_distance(x0,y0,x1,y1):
          d *= -1
          res.append((xa2,ya2)) # going back
        else:
          res.append((xa2,ya2)) # going forward
    i += 1
  return res


def point_on_quadratic_bezier( coords, t):
  """based on info from http://en.wikipedia.org/wiki/B%C3%A9zier_curve"""
  (x0,y0,x1,y1,x2,y2) = coords
  x = (1-t)**2*x0 + 2*t*(1-t)*x1 + t**2*x2
  y = (1-t)**2*y0 + 2*t*(1-t)*y1 + t**2*y2
  return x,y


def quadratic_beziere_to_polyline( point, n=10):
  t = 0
  points = []
  for i in range( n+1):
    points.append( point_on_quadratic_bezier( point, t))
    t += 1.0/n
  return points


def is_point_beween_points_of_line( line, point):
  """this function asserts that the point is on a line and
  computes if it is between the points defining the line"""
  x1, y1, x2, y2 = line
  x, y = point
  if point_distance( x1,y1,x,y) + point_distance( x2,y2,x,y) > 1.02 * point_distance( x1,y1,x2,y2):
    return False
  return True


def plane_normal_from_3_points( point1, point2, point3):
  for point in (point1,point2,point3):
    if None in point:
      return None  # some coords are missing
  from . import transform
  x1,y1,z1 = point1
  x2,y2,z2 = point2
  x3,y3,z3 = point3
  m0 = transform.matrix( [[x1,y1,z1],[x2,y2,z2],[x3,y3,z3]])
  m1 = transform.matrix( [[1,y1,z1],[1,y2,z2],[1,y3,z3]])
  m2 = transform.matrix( [[x1,1,z1],[x2,1,z2],[x3,1,z3]])
  m3 = transform.matrix( [[x1,y1,1],[x2,y2,1],[x3,y3,1]])
  #d0 = 1.0*m0.get_determinant()
  d1 = 1.0*m1.get_determinant()
  d2 = 1.0*m2.get_determinant()
  d3 = 1.0*m3.get_determinant()
  a = d1 #/d0
  b = d2 #/d0
  c = d3 #/d0
  return a,b,c


def angle_between_planes( plane1, plane2):
  a1,b1,c1 = plane1
  a2,b2,c2 = plane2
  cos = (a1*a2 + b1*b2 + c1*c2) / sqrt( a1**2+b1**2+c1**2) / sqrt( a2**2+b2**2+c2**2)
  return cos


def same_or_oposite_side( plane1,plane2):
  a1,b1,c1 = plane1
  a2,b2,c2 = plane2
  a = a1+a2
  b = b1+b2
  c = c1+c2
  if (a**2 + b**2 + c**2) > (a1**2 + b1**2 + c1**2):
    return 1
  else:
    return -1


def expand_rectangle( coords, d):
  x1,y1,x2,y2 = coords
  return x1-d,y1-d,x2+d,y2+d


def create_transformation_to_coincide_point_with_z_axis( mov, point):
  """takes 3d coordinates 'point' (vector mov->point) and returns 3d transform object
  (transform3d.transform3d) that performs rotation to get 'point' onto z axis (x,y)=(0,0)
  with positive 'z'.
  NOTE: this is probably far from efficient, but it works
  """
  from .transform3d import transform3d
  t = transform3d()
  a,b,c = mov
  t.set_move( -a, -b, -c)
  x,y,z = t.transform_xyz( *point)
  t.set_rotation_y( atan2( x, z))
  x,y,z = t.transform_xyz( *point)
  t.set_rotation_x( -atan2( y, sqrt(x**2+z**2)))
  x,y,z = t.transform_xyz( *point)
  if z < 0:
      t.set_rotation_x( pi)
  #t.set_move( *mov)
  return t


def create_transformation_to_rotate_around_particular_axis( line_start, line_end, theta):
  """
  taken from http://inside.mines.edu/~gmurray/ArbitraryAxisRotation/ArbitraryAxisRotation.html
  """
  a,b,c = line_start
  u,v,w = line_end
  u -= a
  v -= b
  w -= c
  u2 = u*u;
  v2 = v*v;
  w2 = w*w;
  cosT = cos(theta);
  sinT = sin(theta);
  l2 = u2 + v2 + w2;
  l =  sqrt(l2);

  if (l2 < 0.000000001):
      raise ValueError("RotationMatrix: direction vector too short!")

  m11 = (u2 + (v2 + w2) * cosT)/l2;
  m12 = (u*v * (1 - cosT) - w*l*sinT)/l2;
  m13 = (u*w * (1 - cosT) + v*l*sinT)/l2;
  m14 = (a*(v2 + w2) - u*(b*v + c*w)
      + (u*(b*v + c*w) - a*(v2 + w2))*cosT + (b*w - c*v)*l*sinT)/l2;

  m21 = (u*v * (1 - cosT) + w*l*sinT)/l2;
  m22 = (v2 + (u2 + w2) * cosT)/l2;
  m23 = (v*w * (1 - cosT) - u*l*sinT)/l2;
  m24 = (b*(u2 + w2) - v*(a*u + c*w)
      + (v*(a*u + c*w) - b*(u2 + w2))*cosT + (c*u - a*w)*l*sinT)/l2;

  m31 = (u*w * (1 - cosT) - v*l*sinT)/l2;
  m32 = (v*w * (1 - cosT) + u*l*sinT)/l2;
  m33 = (w2 + (u2 + v2) * cosT)/l2;
  m34 = (c*(u2 + v2) - w*(a*u + b*v)
      + (w*(a*u + b*v) - c*(u2 + v2))*cosT + (a*v - b*u)*l*sinT)/l2;
  from .transform3d import transform3d
  t = transform3d( [[m11,m12,m13,m14],[m21,m22,m23,m24],[m31,m32,m33,m34],[0,0,0,1]])
  return t




#--------------------------------------------------------------------------
#     This file is part of OASA - a free chemical python library
#     Copyright (C) 2003-2008 Beda Kosata <beda@zirael.org>

#     This program is free software; you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation; either version 2 of the License, or
#     (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.

#     Complete text of GNU GPL can be found in the file gpl.txt in the
#     main directory of the program

#--------------------------------------------------------------------------



class oasa_error( Exception):

  def __init__( self, *args, **kw):
    Exception.__init__( self, *args, **kw)    




class oasa_periodic_table_error( oasa_error):
  """exception for reporting periodic_table related error"""

  def __init__( self, id, value, symbol=None):
    oasa_error.__init__(self)
    self.id = id
    self.value = value

  def __str__( self):
    return "OASA periodic_table error, id=%s, value=%s" % (self.id, self.value)




class oasa_invalid_atom_symbol( oasa_error):
  """exception for reporting invalid atom symbol use"""

  def __init__( self, value, symbol):
    oasa_error.__init__(self)
    self.value = value
    self.symbol = symbol

  def __str__( self):
    return "Symbol '%s' not allowed (%s)" % (self.symbol, self.value)



class oasa_invalid_value( oasa_error):
  """exception for reporting invalid values"""

  def __init__( self, meaning, value):
    oasa_error.__init__(self)
    self.value = value
    self.meaning = meaning

  def __str__( self):
    return "The value for '%s' is not allowed (%s)" % (self.meaning, self.value)



class oasa_not_implemented_error( oasa_error):

  def __init__( self, where, what):
    oasa_error.__init__(self)
    self.where = where
    self.what = what

  def __str__( self):
    return "'Not implemented' error in %s: %s" % (self.where, self.what)


class oasa_inchi_error( oasa_error):

  def __init__( self, what):
    oasa_error.__init__(self)
    self.what = what

  def __str__( self):
    return "InChI error: %s" % self.what



class oasa_unsupported_inchi_version_error( oasa_error):
  
  def __init__( self, version):
    oasa_error.__init__(self)
    self.version = version

  def __str__( self):
    return "The InChI has an unsupported version: %s" % self.version



class oasa_smiles_error( oasa_error):
  
  def __init__( self, value):
    oasa_error.__init__(self)
    self.value = value

  def __str__( self):
    return "SMILES Error: %s" % self.value


class oasa_stereochemistry_error( oasa_error):
  
  def __init__( self, value):
    oasa_error.__init__(self)
    self.value = value

  def __str__( self):
    return "Stereochemistry Error: %s" % self.value






#--------------------------------------------------------------------------
#     This file is part of OASA - a free chemical python library
#     Copyright (C) 2003-2008 Beda Kosata <beda@zirael.org>

#     This program is free software; you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation; either version 2 of the License, or
#     (at your option) any later version.

#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.

#     Complete text of GNU GPL can be found in the file gpl.txt in the
#     main directory of the program

#--------------------------------------------------------------------------



def gen_combinations_of_series( series):
  """series is a list of lists (tuples), the generator yields
  lists by combining each element of each list with each other"""
  counter = len( series) * [0]
  end = [i-1 for i in map( len, series)]
  counter[0] = -1 # dirty trick
  while counter != end:
    for i, e in enumerate( end):
      if counter[i] < e:
        counter[i] += 1
        for j in range( i):
          counter[j] = 0
        break
    yield [s[ counter[ j]] for j,s in enumerate( series)]


def is_uniquely_sorted(series, sorting_function=None):
  """Take a *sorted* series and tell if all the items are unique.

  """
  if sorting_function is None:
    for i in range(len(series) - 1):
      if series[i] == series[i+1]:
        return False
  else:
    for i in range(len(series) - 1):
      if sorting_function(series[i], series[i+1]) == 0:
        return False
  return True


def least_common_item( series):
  d = {}
  for i in series:
    d[i] = d.get( i, 0) + 1
  return list(d.keys())[list(d.values()).index(min(d.values()))]

